"""Command line option handler."""

import numpy as np
from phonopy.cui.settings import ConfParser, Settings


class PhelelSettings(Settings):
    """Setting parameter container."""

    _default = {
        "create_derivatives": None,
        "fft_mesh_numbers": None,
        "finufft_eps": None,
        "grid_points": None,
        "phonon_supercell_matrix": None,
        "subtract_rfs": False,
    }

    def __init__(self, default=None):
        """Init method."""
        super().__init__(default=default)
        self._v.update(PhelelSettings._default.copy())
        if default is not None:
            self._v.update(default)

    def set_create_derivatives(self, val):
        """Setter of create_derivatives."""
        self._v["create_derivatives"] = val

    def set_fft_mesh_numbers(self, val):
        """Setter of fft_mesh_numbers."""
        self._v["fft_mesh_numbers"] = val

    def set_finufft_eps(self, val):
        """Setter of finufft_eps."""
        self._v["finufft_eps"] = val

    def set_grid_points(self, val):
        """Setter of grid_points."""
        self._v["grid_points"] = val

    def set_phonon_supercell_matrix(self, val):
        """Set phonon_supercell_matrix."""
        self._v["phonon_supercell_matrix"] = val

    def set_subtract_rfs(self, val):
        """Setter of subtract_rfs."""
        self._v["subtract_rfs"] = val


class PhelelConfParser(ConfParser):
    """Phelel setting parameter parser."""

    def __init__(self, filename=None, args=None, default_settings=None):
        """Init method."""
        self._settings = PhelelSettings(default=default_settings)
        confs = {}
        if filename is not None:
            super().__init__(filename=filename)
            self.read_file()  # store .conf file setting in self._confs
            self._parse_conf()
            self._set_settings()
            confs.update(self._confs)
        if args is not None:
            super().__init__(args=args)
            self._read_options()
            self._parse_conf()
            self._set_settings()
            confs.update(self._confs)
        self._confs = confs

    def _read_options(self):
        self.read_options()  # store data in self._confs
        if "create_derivatives" in self._args:
            if self._args.create_derivatives:
                dir_names = self._args.create_derivatives
                self._confs["create_derivatives"] = " ".join(dir_names)
        if "fft_mesh_numbers" in self._args:
            if self._args.fft_mesh_numbers:
                self._confs["fft_mesh"] = " ".join(self._args.fft_mesh_numbers)
        if "finufft_eps" in self._args:
            if self._args.finufft_eps is not None:
                self._confs["finufft_eps"] = self._args.finufft_eps
        if "phonon_supercell_dimension" in self._args:
            dim_phonon = self._args.phonon_supercell_dimension
            if dim_phonon is not None:
                self._confs["dim_phonon"] = " ".join(dim_phonon)
        if "subtract_rfs" in self._args:
            if self._args.subtract_rfs:
                self._confs["subtract_rfs"] = ".true."

    def _parse_conf(self):
        self.parse_conf()
        confs = self._confs

        for conf_key in confs.keys():
            if conf_key == "create_derivatives":
                self.set_parameter(
                    "create_derivatives", confs["create_derivatives"].split()
                )

            if conf_key == "dim_phonon":
                matrix = [int(x) for x in confs["dim_phonon"].split()]
                if len(matrix) == 9:
                    matrix = np.array(matrix).reshape(3, 3)
                elif len(matrix) == 3:
                    matrix = np.diag(matrix)
                else:
                    self.setting_error(
                        "Number of elements of dim-phonon has to be 3 or 9."
                    )

                if matrix.shape == (3, 3):
                    if np.linalg.det(matrix) < 1:
                        self.setting_error(
                            "Determinant of supercell matrix has " + "to be positive."
                        )
                    else:
                        self.set_parameter("dim_phonon", matrix)

            if conf_key == "fft_mesh":
                fft_mesh_nums = [int(x) for x in confs["fft_mesh"].split()]
                if len(fft_mesh_nums) == 3:
                    self.set_parameter("fft_mesh_numbers", fft_mesh_nums)
                else:
                    self.setting_error(
                        "Number of elements of fft_mesh tag has to be 3."
                    )

            if conf_key == "finufft_eps":
                self.set_parameter("finufft_eps", confs["finufft_eps"])

            if conf_key == "subtract_rfs":
                if confs["subtract_rfs"] == ".true.":
                    self.set_parameter("subtract_rfs", True)

    def _set_settings(self):
        ConfParser.set_settings(self)
        params = self._parameters

        if "create_derivatives" in params:
            if params["create_derivatives"]:
                self._settings.set_create_derivatives(params["create_derivatives"])

        if "dim_phonon" in params:
            self._settings.set_phonon_supercell_matrix(params["dim_phonon"])

        if "fft_mesh_numbers" in params:
            if params["fft_mesh_numbers"]:
                self._settings.set_fft_mesh_numbers(params["fft_mesh_numbers"])

        if "finufft_eps" in params:
            if params["finufft_eps"]:
                self._settings.set_finufft_eps(params["finufft_eps"])

        if "subtract_rfs" in params:
            if params["subtract_rfs"]:
                self._settings.set_subtract_rfs(params["subtract_rfs"])



"""Implementation of velph-ph_bands-generate."""

import pathlib
import shutil

import click
import tomli
from phonopy.interface.calculator import write_crystal_structure
from phonopy.structure.atoms import parse_cell_dict

from phelel.velph.cli.utils import (
    get_scheduler_dict,
    write_incar,
    write_kpoints_line_mode,
    write_kpoints_mesh_mode,
    write_launch_script,
)


def write_input_files(toml_filename: pathlib.Path) -> None:
    """Generate VASP inputs to generate phonon band structure."""
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    main_directory_name = "ph_bands"
    main_directory = pathlib.Path(main_directory_name)
    main_directory.mkdir(parents=True, exist_ok=True)

    directory_name = f"{main_directory_name}/bands"
    directory = pathlib.Path(directory_name)
    directory.mkdir(parents=True, exist_ok=True)

    # POSCAR
    primitive = parse_cell_dict(toml_dict["primitive_cell"])
    write_crystal_structure(directory / "POSCAR", primitive)

    # INCAR
    write_incar(toml_dict["vasp"]["ph_bands"]["incar"], directory, cell=primitive)

    # KPOINTS
    write_kpoints_mesh_mode(
        toml_dict["vasp"]["ph_bands"]["incar"],
        directory,
        "vasp.ph_bands.kpoints",
        toml_dict["vasp"]["ph_bands"]["kpoints"],
    )

    # QPOINTS
    if "path" in toml_dict["vasp"]["ph_bands"]["qpoints"]:
        click.echo("Seek-path (https://github.com/giovannipizzi/seekpath) is used.")

    write_kpoints_line_mode(
        primitive,
        directory,
        "vasp.ph_bands.qpoints",
        toml_dict["vasp"]["ph_bands"]["qpoints"],
        kpoints_filename="QPOINTS",
    )

    # phelel_params.hdf5
    if pathlib.Path("supercell/phelel_params.hdf5").exists():
        shutil.copy2(
            "supercell/phelel_params.hdf5", "ph_bands/bands/phelel_params.hdf5"
        )
    else:
        click.echo('"supercell/bands/phelel_params.hdf5" not found.', err=True)
        return None

    # POTCAR
    if pathlib.Path("POTCAR").exists():
        shutil.copy2(pathlib.Path("POTCAR"), directory / "POTCAR")

    # Scheduler launch script
    if "scheduler" in toml_dict:
        scheduler_dict = get_scheduler_dict(toml_dict, "ph_bands")
        write_launch_script(scheduler_dict, directory)

    click.echo(f'VASP input files were made in "{directory_name}".')



"""Calculation of DDijQij."""

from __future__ import annotations

from typing import Optional

import numpy as np
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import compute_all_sg_permutations
from phonopy.structure.symmetry import Symmetry

from phelel.base.local_potential import (
    collect_site_symmetry_operations,
    get_displacements_with_rotations,
    rotate_delta_vals_in_spin_space,
)
from phelel.utils.spherical_harmonics import LxLyLzMatrices, SHRotationMatrices


class DeltaDijQij:
    """Container to store delta-Dij and delta-qij.

    Attributes
    ----------
    dDij : ndarray
        Difference of PAW strengths with/without a displacement.
        dtype=complex128
        shape=(ncdij, atoms, lm, lm')
    dqij :
        Difference of <phi_i|phi_j>-<phi_i~|phi_j~>  with/without a
        displacement.
        dtype=complex128
        shape=(ncdij, atoms, lm, lm')
    displacement : dict
        Dispalcement of one atom.
        keys :
            'displacement': Displacement in Cartesian coordinates
            'number': Index of displaced atom
    lm_channels : list of dicts
        lm channels in Dij and qij of atoms
        keys of each distionary:
            'channels' : List of l channels
            'l', 'm' in each channel : l and list of m

    """

    def __init__(
        self,
        Dij_per: np.ndarray,
        Dij_disp: np.ndarray,
        qij_per: np.ndarray,
        qij_disp: np.ndarray,
        displacement: dict,
        lm_channels: list[dict],
    ):
        """Init method.

        Note
        ----
        The order of lm is like
            (l=0,m=0), (l=1,m=-1), (l=1,m=0), (l=1,m=1), (l=2,m=-2),
            (l=2,m=-1), (l=2,m=0), ...

        Parameters
        ----------
        Dij_per : ndarray
            PAW strength of perfect supercell
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        Dij_disp : ndarray
            PAW strength of supercell with a displacement
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        qij_per : ndarray
            <phi_i|phi_j>-<phi_i~|phi_j~> of perfect supercell
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        qij_disp : ndarray
            <phi_i|phi_j>-<phi_i~|phi_j~> of supercell with a displacement
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        displacement : dict
            Dispalcement of one atom
            keys :
                'displacement': Displacement in Cartesian coordinates
                'number': Index of displaced atom
        lm_channels : list of dicts
            lm channels in Dij and qij of atoms
            keys of each distionary:
                'channels' : List of l channels
                'l', 'm' in each channel : l and list of m

        """
        self.dDij = Dij_disp - Dij_per
        self.dqij = qij_disp - qij_per
        self.displacement = displacement
        self.lm_channels = lm_channels


class DDijQijFit:
    """Compute dDij/du and dqij/du for one atom.

    Note
    ----
    Currently ncdij=1.

    Attributes
    ----------
    dDijdu : ndarray
        Derivative of Dij with respect to displacement of an atom
        dtype=complex128
        shape=(ncdij, atom_indices, 3, supercell_atoms, lm, lm')
    dqijdu : ndarray
        Derivative of qij (<phi_i|phi_j>-<phi_i~|phi_j~>) with respect to
        displacement of an atom
        dtype=complex128
        shape=(ncdij, atom_indices, 3, supercell_atoms, lm, lm')
    atom_indices : list of int
        Atom indices in supercell where dV will be computed. If those indices
        don't belong to symmetrically equivalent atoms to the dispalced atom.

    """

    def __init__(
        self,
        delta_Dij_qijs: list[DeltaDijQij],
        supercell: PhonopyAtoms,
        symmetry: Symmetry,
        atom_indices: Optional[list[int]] = None,
        verbose: bool = True,
    ):
        """Init method.

        Note
        ----

        See array shape of Dij and qij at DeltaDijQij class. The private
        vaiables and algorithm related to site symmetry (or site symmetry like)
        are equivalent to those of LocalPotentialInterpolation. A bit more
        detailed comments are written there.

        Parameters
        ----------
        dDij_qijs : list of DeltaDijQij's
            Changes of Dij's and qij's between displaced and perfect supercells
        supercell : PhonopyAtoms
            Perfect supercell
        symmetry : Symmetry
            Symmetry of supercell
        atom_indices : list of int
            Atom indices in supercell where dDij and dqij will be expected to be
            computed. If None, supposed to be all atoms. Internally only
            symmetrically equivalent atoms to the dispalced atom are selected to
            compute.
        verbose : bool
            To display log or not

        """
        # Inputs
        self._delta_Dij_qijs = delta_Dij_qijs
        self._supercell = supercell
        self._symmetry = symmetry
        self._verbose = verbose
        self._atom_indices_in = atom_indices

        self._dDijdu: Optional[np.ndarray] = None
        self._dqijdu: Optional[np.ndarray] = None
        self._atom_indices: Optional[np.ndarray] = None

        # See comments in LocalPotentialInterpolation for these variables
        self._i_atom: Optional[int] = None
        self._sitesym_sets: Optional[np.ndarray] = None

        self._setup()

    @property
    def dDijdu(self):
        """Return dDij/du."""
        return self._dDijdu

    @property
    def dqijdu(self):
        """Return dqij/du."""
        return self._dqijdu

    @property
    def atom_indices(self):
        """Return atom indices where dDijdu and dqijdu are stored."""
        return self._atom_indices

    def __iter__(self):
        """Activate iterator."""
        return self

    def __next__(self):
        """Iterate over atom_indices."""
        if len(self._atom_indices) == self._i_atom:
            raise StopIteration

        if self._verbose:
            print(
                "Computing dDij/du and dqij/du of displaced atom %d:"
                % (self._atom_indices[self._i_atom] + 1)
            )

        self._run_at_atom()
        self._i_atom += 1

    def run(self):
        """Calculate by iterating at atom_indices."""
        for _ in self:
            pass

    def __str__(self):
        """Return Dij and qij in str fashion."""
        text = ""
        Dij_shape = self._dDijdu.shape[-2:]
        for i_spinor, (Dij_spinor, qij_spinor) in enumerate(
            zip(self._dDijdu, self._dqijdu)
        ):
            for i_eatom, (Dij_eatom, qij_eatom) in enumerate(
                zip(Dij_spinor, qij_spinor)
            ):
                for i in range(len(self._supercell)):
                    for j in range(3):
                        text += self._str_format(
                            Dij_eatom[j, i].real,
                            i_spinor,
                            i_eatom,
                            i,
                            j,
                            "D",
                            "real",
                            Dij_shape,
                        )
                        text += self._str_format(
                            qij_eatom[j, i].real,
                            i_spinor,
                            i_eatom,
                            i,
                            j,
                            "q",
                            "real",
                            Dij_shape,
                        )
        return text

    def _str_format(
        self,
        xij: np.ndarray,
        i_spinor: int,
        i_eatom: int,
        i: int,
        j: int,
        str1: str,
        str2: str,
        Dij_shape: tuple,
    ):
        text = (
            f"Spin {i_spinor + 1}, Disp-Atom {i_eatom + 1}, "
            f'Atom {i + 1} d{str1}ij/d{"xyz"[j]} {str2} part '
            f"({Dij_shape[0]}, {Dij_shape[1]})\n"
        )
        for xij_l in xij:
            text += " ".join(["%8.1e" % v for v in xij_l]) + "\n"
        return text

    def _setup(self):
        self._i_atom = 0
        disp_atom = self._delta_Dij_qijs[0].displacement["number"]
        sitesym_sets, equiv_atoms = collect_site_symmetry_operations(
            disp_atom, self._symmetry
        )

        if self._atom_indices_in is None:
            atoms = np.arange(len(self._supercell))
        else:
            atoms = self._atom_indices_in
        self._atom_indices = np.array(
            [i for i in atoms if i in equiv_atoms], dtype="int_"
        )
        sitesym_selected_indices = [
            i for i, eq_atom in enumerate(equiv_atoms) if eq_atom in self._atom_indices
        ]
        self._sitesym_sets = sitesym_sets[sitesym_selected_indices]

        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        natom = len(self._supercell)
        ncdij = self._delta_Dij_qijs[0].dDij.shape[0]
        lmdim = self._delta_Dij_qijs[0].dDij.shape[-2]
        self._dDijdu = np.zeros(
            (ncdij, len(self._atom_indices), 3, natom, lmdim, lmdim),
            dtype=dtype,
            order="C",
        )
        self._dqijdu = np.zeros(
            (ncdij, len(self._atom_indices), 3, natom, lmdim, lmdim),
            dtype=dtype,
            order="C",
        )

    def _run_at_atom(self):
        sitesyms = self._sitesym_sets[self._i_atom]
        natom = len(self._supercell)
        lmdim = self._delta_Dij_qijs[0].dDij.shape[-2]
        dtype = f'c{np.dtype("double").itemsize * 2}'
        ncdij = self._dDijdu.shape[0]

        lattice = self._supercell.cell.T
        rotations = self._symmetry.symmetry_operations["rotations"][sitesyms]
        translations = self._symmetry.symmetry_operations["translations"][sitesyms]
        atomic_permutations = compute_all_sg_permutations(
            self._supercell.scaled_positions,
            rotations,
            translations,
            np.array(lattice, dtype="double", order="C"),
            self._symmetry.tolerance,
        )

        disps = get_displacements_with_rotations(
            rotations, lattice, self._delta_Dij_qijs
        )
        disps_inv = np.linalg.pinv(disps)
        dDij_rotated_all = np.zeros(
            (ncdij, len(disps), natom * lmdim**2), dtype=dtype, order="C"
        )
        dqij_rotated_all = np.zeros(
            (ncdij, len(disps), natom * lmdim**2), dtype=dtype, order="C"
        )

        count = 0
        for delta_Dij_qij in self._delta_Dij_qijs:
            for r, perm in zip(rotations, atomic_permutations):
                dDij_rotated, dqij_rotated = self._rotate_Dij_qij(
                    ncdij, delta_Dij_qij, perm, r
                )
                if ncdij == 4:  # Need to rotate in spin space, too.
                    dDij_spinor_rotated = rotate_delta_vals_in_spin_space(
                        dDij_rotated, r, lattice
                    )
                    dqij_spinor_rotated = rotate_delta_vals_in_spin_space(
                        dqij_rotated, r, lattice
                    )
                    dDij_rotated = dDij_spinor_rotated
                    dqij_rotated = dqij_spinor_rotated
                for i_cdij in range(ncdij):
                    dDij_rotated_all[i_cdij, count] = dDij_rotated[i_cdij]
                    dqij_rotated_all[i_cdij, count] = dqij_rotated[i_cdij]
                count += 1

        # Compute dDij/du and dqij/du
        # shape=(ncdij, len(self._atom_indices), 3, natom, lmdim, lmdim)
        shape = (ncdij, 3, natom, lmdim, lmdim)
        self._dDijdu[:, self._i_atom] = (disps_inv @ dDij_rotated_all).reshape(shape)
        self._dqijdu[:, self._i_atom] = (disps_inv @ dqij_rotated_all).reshape(shape)

    def _rotate_Dij_qij(
        self, ncdij: int, delta_Dij_qij: DeltaDijQij, perm: np.ndarray, r: np.ndarray
    ) -> list[np.ndarray, np.ndarray]:
        """Rotate Dij and qij.

        This rotation is the direct product of rotations of atomic permutation
        and atomic-like orbitals on atomic points. The displacements are rotated
        actively (R), and atomic permutation and atomic-like orbitals are
        rotated passively (R^-1).

        """
        rot_dDij = []
        rot_dqij = []
        for i_cdij in range(ncdij):
            perm_inv = np.zeros_like(perm)
            for j, k in enumerate(perm):
                perm_inv[k] = j
            _rot_dDij, _rot_dqij = self._get_inv_rotated_dDij_qij(
                delta_Dij_qij, r, i_cdij
            )
            rot_dDij.append(_rot_dDij[perm_inv].ravel())
            rot_dqij.append(_rot_dqij[perm_inv].ravel())
        return rot_dDij, rot_dqij

    def _get_inv_rotated_dDij_qij(self, delta_Dij_qij: DeltaDijQij, r, i_spinor):
        """Inverse-rotate dDij and dqij."""
        lattice = self._supercell.cell.T
        shr = SHRotationMatrices(r, lattice, LxLyLzMatrices().run())
        shr.run()

        rot_dDij = np.zeros_like(delta_Dij_qij.dDij[0])
        rot_dqij = np.zeros_like(delta_Dij_qij.dqij[0])

        # loop over atoms
        for i, (dDij, dqij, lm_channels) in enumerate(
            zip(
                delta_Dij_qij.dDij[i_spinor],
                delta_Dij_qij.dqij[i_spinor],
                delta_Dij_qij.lm_channels,
            )
        ):
            bigDelta = self._get_big_Delta(shr.Delta, lm_channels, dDij)
            rot_dDij[i] = bigDelta @ dDij @ bigDelta.T.conj()
            rot_dqij[i] = bigDelta @ dqij @ bigDelta.T.conj()

        return rot_dDij, rot_dqij

    def _get_big_Delta(self, Delta, lm_channels, dDij):
        """Return orbital rotation matrix.

        This matrix combines rotation matrices of different orbitals, and
        is a unitary matrix.

        """
        bigDelta = np.zeros_like(dDij)
        row = 0
        if lm_channels["channels"] is None:
            return bigDelta
        for ll in lm_channels["channels"]:
            n = ll["l"] * 2 + 1
            bigDelta[row : (row + n), row : (row + n)] = Delta[ll["l"]]
            row += n
        return bigDelta


class DDijQij:
    """Compute dDij/du and dqij/du.

    Note
    ----
    ncdij = 1 without spin
    ncdij = 2 with collinear spin (up and down)
    ncdij = 4 with non-collinear spin (2x2 localpotential)

    Attributes
    ----------
    dDijdu : ndarray
        Derivative of Dij with respect to displacements
        dtype=complex128
        shape=(ncdij, atom_indices, 3, supercell_atoms, lm, lm')
    dqijdu : ndarray
        Derivative of qij (<phi_i|phi_j>-<phi_i~|phi_j~>) with respect to
        displacements
        dtype=complex128
        shape=(ncdij, atom_indices, 3, supercell_atoms, lm, lm')
    lm_channels : list of dicts
        lm channels in Dij and qij of atoms
        keys of each distionary:
            'channels' : List of l channels
            'l', 'm' in each channel : l and list of m
    symmetry : Symmetry
        Symmetry of supercell
    atom_indices : ndarray
        Atom indices in supercell where dV is computed. This is made as
        np.unique(atom_indices given at __init__). If None, all atoms in
        supercell.
        shape=(len(atom_indices),)
        dtype='int_'

    """

    def __init__(
        self,
        supercell: PhonopyAtoms,
        symmetry: Optional[Symmetry] = None,
        atom_indices=None,
        verbose=True,
    ):
        """Init method.

        Parameters
        ----------
        supercell : PhonopyAtoms
            Supercell
        symmetry : Symmetry, optional
            Symmetry of supercell. If None, symmetry is searched in this class
            object.
        atom_indices : list of int, optional
            Atom indices in supercell where dDijdu will be expected to be
            computed. If None, supposed to be all atoms. Internally only
            symmetrically equivalent atoms to the dispalced atom are selected
            to compute.

        """
        self._supercell = supercell
        self._verbose = verbose

        if atom_indices is None:
            self.atom_indices = np.arange(len(self._supercell), dtype="int_")
        else:
            self.atom_indices = np.array(np.unique(atom_indices), dtype="int_")
        if symmetry is None:
            self.symmetry = Symmetry(supercell)
        else:
            self.symmetry = symmetry

        self._Dij = None
        self._qij = None
        self._dDijdu = None
        self._dqijdu = None

    @property
    def dDijdu(self):
        """Getter and setter of dDijdu."""
        return self._dDijdu

    @dDijdu.setter
    def dDijdu(self, dDijdu):
        natom = len(self._supercell)
        shape = (len(self.atom_indices), 3, natom)
        if dDijdu.shape[1:4] == shape:
            dtype = "c%d" % (np.dtype("double").itemsize * 2)
            self._dDijdu = np.array(dDijdu, dtype=dtype, order="C")
        else:
            raise RuntimeError(
                "Array shape[1:4] disagreement is found, %s!=%s."
                % (shape, dDijdu.shape[1:4])
            )

    @property
    def dqijdu(self):
        """Getter and setter of dqijdu."""
        return self._dqijdu

    @dqijdu.setter
    def dqijdu(self, dqijdu):
        natom = len(self._supercell)
        shape = (len(self.atom_indices), 3, natom)
        if dqijdu.shape[1:4] == shape:
            dtype = "c%d" % (np.dtype("double").itemsize * 2)
            self._dqijdu = np.array(dqijdu, dtype=dtype, order="C")
        else:
            raise RuntimeError(
                "Array shape[1:4] disagreement is found, %s!=%s."
                % (shape, dqijdu.shape[1:4])
            )

    @property
    def Dij(self):
        """Getter and setter of Dij."""
        return self._Dij

    @Dij.setter
    def Dij(self, Dij):
        if Dij.shape[1] == len(self.atom_indices):
            dtype = "c%d" % (np.dtype("double").itemsize * 2)
            self._Dij = np.array(Dij, dtype=dtype, order="C")
        else:
            raise RuntimeError(
                "Number of atoms in supercell disagrees, %s!=%s."
                % (len(self.atom_indices), Dij.shape[1])
            )

    @property
    def qij(self):
        """Getter and setter of qij."""
        return self._qij

    @qij.setter
    def qij(self, qij):
        if qij.shape[1] == len(self.atom_indices):
            dtype = "c%d" % (np.dtype("double").itemsize * 2)
            self._qij = np.array(qij, dtype=dtype, order="C")
        else:
            raise RuntimeError(
                "Number of atoms in supercell disagrees, %s!=%s."
                % (len(self.atom_indices), qij.shape[1])
            )

    def run(self, Dij_per, Dij_disps, qij_per, qij_disps, displacements, lm_channels):
        """Compute dDij/du and dqij/du.

        Parameters
        ----------
        Dij_per : ndarray
            PAW strength of perfect supercell
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        Dij_disp : ndarray
            PAW strength of supercell with a displacement
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        qij_per : ndarray
            <phi_i|phi_j>-<phi_i~|phi_j~> of perfect supercell
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        qij_disp : ndarray
            <phi_i|phi_j>-<phi_i~|phi_j~> of supercell with a displacement
            dtype=complex128
            shape=(ncdij, atoms, lm, lm')
        displacement : dict
            Dispalcement of one atom
            keys :
                'displacement': Displacement in Cartesian coordinates
                'number': Index of displaced atom
        lm_channels : list of dicts
            lm channels in Dij and qij of atoms
            keys of each distionary:
                'channels' : List of l channels
                'l', 'm' in each channel : l and list of m

        """
        if self.dDijdu is None:
            self._allocate_arrays(Dij_per.shape[0], Dij_per.shape[2])

        for disp_atom in np.unique([d["number"] for d in displacements]):
            delta_Dij_qijs = []
            for i, d in enumerate(displacements):
                if d["number"] == disp_atom:
                    delta_Dij_qijs.append(
                        DeltaDijQij(
                            Dij_per,
                            Dij_disps[i],
                            qij_per,
                            qij_disps[i],
                            displacements[i],
                            lm_channels,
                        )
                    )
            ddijqij = DDijQijFit(
                delta_Dij_qijs,
                self._supercell,
                self.symmetry,
                atom_indices=self.atom_indices,
                verbose=self._verbose,
            )
            ddijqij.run()

            indices = []
            for ai in ddijqij._atom_indices:
                indices.append(np.where(self.atom_indices == ai)[0][0])
            self._dDijdu[:, indices] = ddijqij._dDijdu
            self._dqijdu[:, indices] = ddijqij._dqijdu

        self.Dij = Dij_per[:, self.atom_indices, :, :]
        self.qij = qij_per[:, self.atom_indices, :, :]

    def _allocate_arrays(self, ncdij, lmdim):
        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        natom = len(self._supercell)
        shape = (ncdij, len(self.atom_indices), 3, natom, lmdim, lmdim)
        self._dDijdu = np.zeros(shape, dtype=dtype, order="C")
        self._dqijdu = np.zeros(shape, dtype=dtype, order="C")



"""Implementation of velph-phelel-differentiate."""

import pathlib

import click
import tomli

import phelel
from phelel.interface.vasp.derivatives import read_forces_from_vasprunxmls
from phelel.velph.cli.utils import get_num_digits


def create_phonopy_yaml(
    toml_filename: pathlib.Path,
    yaml_filename: pathlib.Path,
    dir_name: str,
    subtract_residual_forces: bool = True,
):
    """Calculate derivatives and write phelel_params.hdf5."""
    phonopy_yaml_filename = pathlib.Path(f"{dir_name}/phonopy_params.yaml")

    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    is_symmetry = True
    try:
        if toml_dict["phelel"]["nosym"] is True:
            click.echo(
                'Found "nosym = true" in [phelel] section. '
                "Symmetrization is turned off."
            )
            is_symmetry = False
    except KeyError:
        pass

    phe = phelel.load(
        str(yaml_filename),
        is_symmetry=is_symmetry,
    )
    dir_names = []
    nd = get_num_digits(phe.supercells_with_displacements)
    for i, _ in enumerate(
        [
            phe.supercell,
        ]
        + phe.supercells_with_displacements
    ):
        id_number = f"{i:0{nd}d}"
        dir_names.append(pathlib.Path(f"{dir_name}/disp-{id_number}"))

    if phe.phonon_supercell_matrix is not None:
        nd = get_num_digits(phe.phonon_supercells_with_displacements)
        for i, _ in enumerate(
            [
                phe.phonon_supercell,
            ]
            + phe.phonon_supercells_with_displacements
        ):
            id_number = f"{i:0{nd}d}"
            dir_names.append(pathlib.Path(f"{phelel}/ph-disp-{id_number}"))

    phonopy_yaml_filename.parent.mkdir(parents=True, exist_ok=True)

    # NAC params should be contained phelel_disp.yaml.
    # Therefore nac_params is not set to phe here.
    forces = read_forces_from_vasprunxmls(
        [d / "vasprun.xml" for d in dir_names],
        phe.phonon,
        subtract_rfs=subtract_residual_forces,
        log_level=0,
    )
    phe.phonon.forces = forces
    phe.phonon.save(filename=phonopy_yaml_filename)
    click.echo(f'"{phonopy_yaml_filename}" has been made.')



"""VASP file IO functions."""

import os
from typing import Union

import h5py
import numpy as np
import yaml
from phono3py.file_IO import get_filename_suffix
from phonopy.file_IO import get_io_module_to_decompress


###########
# Readers #
###########
def read_bin_stream(filename, dtype=None):
    """Read binary stream."""
    if dtype is None:
        dtype = np.dtype("double")
    myio = get_io_module_to_decompress(filename)
    with myio.open(filename, "rb") as fp:
        data = np.frombuffer(fp.read(), dtype=dtype)
    return data


def read_inwap_vaspouth5(
    filename: Union[str, bytes, os.PathLike] = "vaspout.h5",
) -> dict:
    """Read inwap-like information in vaspout.h5."""
    inwap = {}
    with h5py.File(filename) as h5:
        # dimensions of the potential
        pot = h5["results/potential/total"]
        ncdij = pot.shape[0]
        fft_fine = pot.shape[1:]
        # electronic structure
        ispin = h5["/results/electron_eigenvalues/ispin"][()]
        nbtot = h5["/results/electron_eigenvalues/nb_tot"][()]
        nkpts = h5["/results/electron_eigenvalues/kpoints"][()]
        # number of ions
        nions = h5["/results/electron_eigenvalues/nions"][()]
        # kpoint coordinates
        kpoints = h5["/results/electron_eigenvalues/kpoint_coords"][:]
        inwap["ncdij"] = ncdij
        inwap["nions"] = nions
        inwap["fft_fine"] = fft_fine
        inwap["ispin"] = ispin
        inwap["nkpts"] = nkpts
        inwap["nbtot"] = nbtot
        inwap["nrspinors"] = 2 if ncdij == 4 else 1
        inwap["kpoints"] = kpoints
        lmdim = h5["results/paw/lmdim"][()]
        ldim = h5["results/paw/ldim"][0]
        inwap["lmdim"] = lmdim  # data["PAW"]["LMDIM"]
        inwap["ldim"] = ldim  # data["PAW"]["LDIM"]

        # read required information to generate lm_orbitals
        nitypes = h5["/input/poscar/number_ion_types"][:]
        lmax = h5["/results/paw/lmax"][:]
        lps = h5["/results/paw/lps"][:]

        # construct lm_orbitals dictionary here
        itypes = sum([[i] * nityp for i, nityp in enumerate(nitypes)], [])
        lm_orbitals = []
        for i in range(nions):
            ityp = itypes[i]
            channels = []
            for il in range(lmax[ityp]):
                ll = lps[ityp, il]
                channels.append({"l": ll, "m": list(range(-ll, ll + 1))})
            atom_dict = {
                "atom_index": i + 1,
                "num_l_channels": lmax[ityp],
                "channels": channels,
            }
            lm_orbitals.append(atom_dict)
        inwap["lm_orbitals"] = lm_orbitals
        # TODO
        inwap["fft_coarse"] = None  # data["fft_grid"]["coarse"]

    return inwap


def read_inwap_yaml(filename: Union[str, bytes, os.PathLike] = "inwap.yaml") -> dict:
    """Read inwap.yaml."""
    inwap = {}
    with open(filename) as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
        inwap["ncdij"] = data["cell"]["NCDIJ"]
        inwap["nions"] = data["cell"]["NIONS"]
        inwap["fft_coarse"] = data["fft_grid"]["coarse"]
        inwap["fft_fine"] = data["fft_grid"]["fine"]
        inwap["ispin"] = data["cell"]["ISPIN"]
        inwap["nkpts"] = data["cell"]["NKPTS"]
        inwap["nbtot"] = data["cell"]["NBTOT"]
        inwap["nrspinors"] = data["cell"]["NRSPINORS"]
        inwap["lmdim"] = data["PAW"]["LMDIM"]
        inwap["ldim"] = data["PAW"]["LDIM"]
        inwap["lm_orbitals"] = data["PAW"]["lm_orbitals"]
        inwap["kpoints"] = data["KPOINTS"]

    return inwap


def read_PAW_Dij_qij_vaspouth5(filename):
    """Read Dij and qij in vaspout.h5.

    Dij(ncdij, nions, lmdim, lmdim')
    qij(ncdij, nions, lmdim, lmdim')

    Used such as <psi|p><lm|A|lm'><p|psi'>

    """
    with h5py.File(filename) as h5:
        dij_real = h5["/results/paw/dij"][:]
        qij_real = h5["/results/paw/qij"][:]
    dij = dij_real[:, :, :, :, 0] + 1j * dij_real[:, :, :, :, 1]
    qij = qij_real[:, :, :, :, 0] + 1j * qij_real[:, :, :, :, 1]
    return dij, qij


def read_PAW_Dij_qij(inwap, filename, is_Rij=False):
    """Read Dij, qij, and Rij.

    Dij(ncdij, nions, lmdim, lmdim')
    qij(ncdij, nions, lmdim, lmdim')
    Rij(2, ncdij, nions, lmdim, lmdim')

    Used such as <psi|p><lm|A|lm'><p|psi'>

    """
    ncdij = inwap["ncdij"]
    nions = inwap["nions"]
    lmdim = inwap["lmdim"]
    if is_Rij:
        shape = (2, ncdij, nions, lmdim, lmdim)
    else:
        shape = (ncdij, nions, lmdim, lmdim)
    data = read_bin_stream(filename)
    dt = "c%d" % (data.itemsize * 2)
    data_complex = data.view(dtype=dt)
    if len(data_complex) != np.prod(shape):
        print("%s: data size is inconsistent with values in inwap.yaml." % filename)
        return None
    return data_complex.reshape(shape)


def read_local_potential_vaspouth5(inwap, filename="vaspout.h5"):
    """Read local potentials in vaspout.h5.

    For spin unpolarized calculations the potential is real but we convert it to
    complex.
    For collinear calculations the potential is written as up and down.
    For non-collinear calculations the potential is written as scalar potential
    + magnetic field potential

    """
    with h5py.File(filename) as h5:
        # dimensions of the potential
        pot_real = h5["/results/potential/total"][:]
        ncdij = pot_real.shape[0]
        pot = np.zeros_like(pot_real, dtype=complex)

        if ncdij == 1:
            v = pot_real[0, :, :, :]
            pot[:, :, :, :] = v
        if ncdij == 2:
            v = pot_real[0, :, :, :]
            vz = pot_real[1, :, :, :]
            pot[0, :, :, :] = v + vz
            pot[1, :, :, :] = v - vz

        if ncdij == 4:
            v = pot_real[0, :, :, :]
            vx = pot_real[1, :, :, :]
            vy = pot_real[2, :, :, :]
            vz = pot_real[3, :, :, :]
            pot[0, :, :, :] = v + vz
            pot[1, :, :, :] = vx - 1j * vy
            pot[2, :, :, :] = vx + 1j * vy
            pot[3, :, :, :] = v - vz
        return pot


def read_local_potential(inwap, filename="LOCAL-POTENTIAL.bin"):
    """Read LOCAL-POTENTIAL.bin.

    Normally local potential (SV) is real in VASP. But non-collinear version, it
    turns to be complex. So we force local potential to be complex in either
    case.

    """
    ncdij = inwap["ncdij"]
    nx, ny, nz = inwap["fft_fine"]
    shape = (ncdij, nz, ny, nx)
    data = read_bin_stream(filename)
    dt = "c%d" % (data.itemsize * 2)
    data_complex = data.view(dtype=dt)
    if len(data_complex) != np.prod(shape):
        print("%s: data size is inconsistent with values in inwap.yaml." % filename)
        return None
    return data_complex.reshape(shape)


def read_dprojectors(inwap, filename="DPROJECTORS.bin"):
    """Read DPROJECTORS.bin."""
    nbtot = inwap["nbtot"]
    nrspinors = inwap["nrspinors"]
    ispin = inwap["ispin"]
    nkpts = inwap["nkpts"]
    lmdim = inwap["lmdim"]
    nions = inwap["nions"]
    # The last index corresponds to usual projector and its derivatives along
    # x, y, z.
    shape = (nkpts, nbtot, ispin, nrspinors, nions, lmdim, 4)
    data = read_bin_stream(filename)
    dt = "c%d" % (data.itemsize * 2)
    data_complex = data.view(dtype=dt)
    if len(data_complex) != np.prod(shape):
        print("%s: data size is inconsistent with values in inwap.yaml." % filename)
        return None
    return data_complex.reshape(shape)


def read_waves(inwap, filename="WAVES.bin"):
    """Read WAVES.bin."""
    ispin = inwap["ispin"]
    nkpts = inwap["nkpts"]
    nbtot = inwap["nbtot"]
    nrspinors = inwap["nrspinors"]
    nx, ny, nz = inwap["fft_coarse"]
    shape = (nkpts, nbtot, ispin, nrspinors, nz, ny, nx)
    data = read_bin_stream(filename)
    dt = "c%d" % (data.itemsize * 2)
    data_complex = data.view(dtype=dt)
    if len(data_complex) != np.prod(shape):
        print("%s: data size is inconsistent with values in inwap.yaml." % filename)
        return None
    return data_complex.reshape(shape)


def read_eigenvalues(inwap, filename="EIGENVALUE.bin"):
    """Read EIGENVALUE.bin."""
    ispin = inwap["ispin"]
    nkpts = inwap["nkpts"]
    nbtot = inwap["nbtot"]
    shape = (nbtot, nkpts, ispin)
    data = read_bin_stream(filename)
    if len(data) != np.prod(shape):
        print("%s: data size is inconsistent with values in inwap.yaml." % filename)
        return None

    # change shape (nbtot, nkpts, ispin) --> (nkpts, nbtot, ispin)
    ret_data = np.array(data.reshape(shape).swapaxes(0, 1), dtype=data.dtype, order="C")
    return ret_data


def read_qtot(inwap, filename="QTOT.bin"):
    """Read QTOT.bin."""
    ldim = inwap["ldim"]
    nions = inwap["nions"]
    shape = (nions, ldim, ldim)
    data = read_bin_stream(filename)
    if len(data) != np.prod(shape):
        print("%s: data size is inconsistent with values in inwap.yaml." % filename)
        return None
    return data.reshape(shape)


def get_CHGCAR(charge, header):
    """Return CHGCAR style text from ndarray."""
    text = header + "\n%5d%5d%5d\n" % charge.shape[::-1]
    count = 0
    for z in range(charge.shape[0]):
        for y in range(charge.shape[1]):
            for x in range(charge.shape[2]):
                text += "%18.11e" % charge[z, y, x]
                count += 1
                if (count % 5) == 0:
                    count = 0
                    text += "\n"
    return text


###########
# Writers #
###########
def write_mesh_electron_hdf5(dirnames, mesh):
    """Read .bin files and write to hdf5 file.

    Note
    ----
    No spin and spinor are considered.

    """
    inwaps = []
    for dname in dirnames:
        inwaps.append(read_inwap_yaml("%s/%s" % (dname, "inwap.yaml")))

    # check all have the same data sizes except for nkpts.
    assert len(np.unique([iwp["nbtot"] for iwp in inwaps])) == 1
    assert len(np.unique([iwp["nions"] for iwp in inwaps])) == 1
    assert len(np.unique([iwp["lmdim"] for iwp in inwaps])) == 1
    fft_meshes = np.array([iwp["fft_coarse"] for iwp in inwaps])
    assert (fft_meshes[1:] - fft_meshes[0] == 0).all()

    nkpts = np.array([iwp["nkpts"] for iwp in inwaps], dtype=int)
    nbtot = inwaps[0]["nbtot"]
    lmdim = inwaps[0]["lmdim"]
    nions = inwaps[0]["nions"]
    fft_mesh = tuple(fft_meshes[0])
    dtype = "c%d" % (np.dtype("double").itemsize * 2)
    # waves = np.empty((np.sum(nkpts), nbtot, 1, 1) + fft_mesh,
    #                  dtype=dtype)
    waves_shape = (np.sum(nkpts), nbtot, 1, 1) + fft_mesh
    dprojs = np.empty((np.sum(nkpts), nbtot, 1, 1, nions, lmdim, 4), dtype=dtype)
    eigenvalues = np.empty((np.sum(nkpts), nbtot, 1), dtype="double")
    idx = 0
    for dname, iwp, nk in zip(dirnames, inwaps, nkpts):
        # waves[idx:(idx + nk)] = read_waves(
        #     iwp, "%s/%s" % (dname, "WAVES.bin"))
        dprojs[idx : (idx + nk)] = read_dprojectors(
            iwp, "%s/%s" % (dname, "DPROJECTORS.bin")
        )
        eigvals = read_eigenvalues(iwp, "%s/%s" % (dname, "EIGENVALUE.bin"))
        eigenvalues[idx : (idx + nk)] = eigvals
        idx += nk

    suffix = get_filename_suffix(mesh)
    with h5py.File("electron%s.hdf5" % suffix, "w") as w:
        # compression doesn't help much.
        w.create_dataset("mesh", data=mesh)
        # w.create_dataset('waves', data=waves, compression=None)
        w.create_dataset("dprojectors", data=dprojs, compression=None)
        w.create_dataset("eigenvalues", data=eigenvalues, compression=None)

    # This is to avoid using memory space by directly writing to the file
    # object piece by piece.
    with h5py.File("electron%s.hdf5" % suffix, "a") as w:
        waves = w.create_dataset("waves", waves_shape, dtype=dtype)
        idx = 0
        for dname, iwp, nk in zip(dirnames, inwaps, nkpts):
            waves[idx : (idx + nk)] = read_waves(iwp, "%s/%s" % (dname, "WAVES.bin"))
            idx += nk


def write_mesh_KPOINTS(filename, num_per_batch=100):
    """Read grid_address-xxx.hdf5 and write KPOINTS style files.

    The set of k-points is divided into kpoint batches and
    each batch has num_per_batch kpoints.

    """
    with h5py.File(filename, "r") as f:
        mesh = np.array(f["mesh"][:], dtype=float)
        grid_address = f["grid_address"][:]

    num = num_per_batch
    w = None
    for i, kpt in enumerate(grid_address / mesh):
        if i % num == 0:
            if i != 0:
                w.close()
            w = open("KPOINTS-%03d" % (i // num + 1), "w")
            if len(grid_address) - i > num:
                w.write("batch %d-%d\n" % (i + 1, i + num))
                w.write("%d\n" % num)
            else:
                w.write("batch %d-%d\n" % (i + 1, len(grid_address)))
                w.write("%d\n" % (len(grid_address) - i))
            w.write("Reciprocal\n")
        w.write("%12.8f%12.8f%12.8f 1\n" % tuple(kpt))
    w.close()



"""phelel_yaml reader and writer."""

# Copyright (C) 2021 Atsushi Togo
# All rights reserved.
#
# This file is part of phelel.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dataclasses
from typing import TYPE_CHECKING, Optional

import numpy as np
from phonopy.interface.phonopy_yaml import (
    PhonopyYaml,
    PhonopyYamlDumper,
    PhonopyYamlLoader,
    load_yaml,
    phonopy_yaml_property_factory,
)
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive, Supercell, isclose
from phonopy.structure.symmetry import Symmetry

if TYPE_CHECKING:
    from phelel import Phelel


@dataclasses.dataclass
class PhelelYamlData:
    """PhonopyYaml data structure."""

    configuration: Optional[dict] = None
    calculator: Optional[str] = None
    physical_units: Optional[dict] = None
    unitcell: Optional[PhonopyAtoms] = None
    primitive: Optional[Primitive] = None
    supercell: Optional[Supercell] = None
    dataset: Optional[dict] = None
    supercell_matrix: Optional[np.ndarray] = None
    primitive_matrix: Optional[np.ndarray] = None
    nac_params: Optional[dict] = None
    force_constants: Optional[np.ndarray] = None
    symmetry: Optional[Symmetry] = None  # symmetry of supercell
    frequency_unit_conversion_factor: Optional[float] = None
    version: Optional[str] = None
    command_name: str = "phelel"

    phonon_supercell_matrix: Optional[np.ndarray] = None
    phonon_dataset: Optional[dict] = None
    phonon_supercell: Optional[Supercell] = None
    phonon_primitive: Optional[Primitive] = None


class PhelelYamlLoader(PhonopyYamlLoader):
    """PhelelYaml loader."""

    def __init__(
        self, yaml_data, configuration=None, calculator=None, physical_units=None
    ):
        """Init method.

        Parameters
        ----------
        yaml_data : dict

        """
        self._yaml = yaml_data
        self._data = PhelelYamlData(
            configuration=configuration,
            calculator=calculator,
            physical_units=physical_units,
        )

    def parse(self):
        """Yaml dict is parsed. See docstring of this class."""
        super().parse()
        self._parse_phonon_dataset()
        return self

    def _parse_all_cells(self):
        """Parse all cells.

        This method override PhonopyYaml._parse_all_cells.

        """
        super()._parse_all_cells()
        if "phonon_primitive_cell" in self._yaml:
            self._data.phonon_primitive = self._parse_cell(
                self._yaml["phonon_primitive_cell"]
            )
        if "phonon_supercell" in self._yaml:
            self._data.phonon_supercell = self._parse_cell(
                self._yaml["phonon_supercell"]
            )
        if "phonon_supercell_matrix" in self._yaml:
            self._data.phonon_supercell_matrix = np.array(
                self._yaml["phonon_supercell_matrix"], dtype="intc", order="C"
            )

    def _parse_phonon_dataset(self):
        """Parse force dataset for phonon."""
        self._data.phonon_dataset = self._get_dataset(
            self._data.phonon_supercell, key_prefix="phonon_"
        )


class PhelelYamlDumper(PhonopyYamlDumper):
    """PhelelYaml dumper."""

    default_settings = {
        "force_sets": False,
        "displacements": True,
        "force_constants": False,
        "born_effective_charge": True,
        "dielectric_constant": True,
    }

    def __init__(self, data: PhelelYamlData, dumper_settings=None):
        """Init method."""
        self._data = data
        self._init_dumper_settings(dumper_settings)

    def _cell_info_yaml_lines(self):
        """Get YAML lines for information of cells.

        This method override PhonopyYaml._cell_info_yaml_lines.

        """
        lines = super()._cell_info_yaml_lines()
        lines += self._supercell_matrix_yaml_lines(
            self._data.phonon_supercell_matrix, "phonon_supercell_matrix"
        )
        lines += self._primitive_yaml_lines(
            self._data.phonon_primitive, "phonon_primitive_cell"
        )
        lines += self._phonon_supercell_yaml_lines()
        return lines

    def _phonon_supercell_matrix_yaml_lines(self):
        lines = []
        if self._data.phonon_supercell_matrix is not None:
            lines.append("phonon_supercell_matrix:")
            for v in self._data.supercell_matrix:
                lines.append("- [ %3d, %3d, %3d ]" % tuple(v))
            lines.append("")
        return lines

    def _phonon_supercell_yaml_lines(self):
        lines = []
        if self._data.phonon_supercell is not None:
            s2p_map = getattr(self._data.phonon_primitive, "s2p_map", None)
            lines += self._cell_yaml_lines(
                self._data.phonon_supercell, "phonon_supercell", s2p_map
            )
            lines.append("")
        return lines

    def _nac_yaml_lines(self):
        """Get YAML lines for parameters of non-analytical term correction.

        This method override PhonopyYaml._nac_yaml_lines.

        """
        assert isclose(self._data.primitive, self._data.phonon_primitive)
        return super()._nac_yaml_lines()

    def _displacements_yaml_lines(self, with_forces=False):
        """Get YAML lines for phonon_dataset and dataset.

        This method override PhonopyYaml._displacements_yaml_lines.
        PhonopyYaml._displacements_yaml_lines_2types is written
        to be also used by Phono3pyYaml.

        """
        lines = []
        if self._data.phonon_supercell_matrix is not None:
            lines += self._displacements_yaml_lines_2types(
                self._data.phonon_dataset,
                with_forces=with_forces,
                key_prefix="phonon_",
            )
        lines += self._displacements_yaml_lines_2types(
            self._data.dataset, with_forces=with_forces
        )
        return lines


class PhelelYaml(PhonopyYaml):
    """phelel.yaml reader and writer.

    Details are found in the docstring of PhonopyYaml.
    The common usages are as follows:

    1. Set phelel instance.
        phe_yml = PhelelYaml()
        phe_yml.set_phelel_info(phelel_instance)
    2. Read phelel.yaml file.
        phe_yml = PhelelYaml()
        phe_yml.read(filename)
    3. Parse yaml dict of phelel.yaml.
        with open("phelel.yaml", 'r') as f:
            phe_yml.yaml_data = yaml.load(f, Loader=yaml.CLoader)
            phe_yml.parse()
    4. Save stored data in PhelelYaml instance into a text file in yaml.
        with open(filename, 'w') as w:
            w.write(str(phe_yml))

    """

    default_filenames = ("phelel_disp.yaml", "phelel.yaml")
    command_name = "phelel"

    configuration = phonopy_yaml_property_factory("configuration")
    calculator = phonopy_yaml_property_factory("calculator")
    physical_units = phonopy_yaml_property_factory("physical_units")
    unitcell = phonopy_yaml_property_factory("unitcell")
    primitive = phonopy_yaml_property_factory("primitive")
    supercell = phonopy_yaml_property_factory("supercell")
    dataset = phonopy_yaml_property_factory("dataset")
    supercell_matrix = phonopy_yaml_property_factory("supercell_matrix")
    primitive_matrix = phonopy_yaml_property_factory("primitive_matrix")
    nac_params = phonopy_yaml_property_factory("nac_params")
    force_constants = phonopy_yaml_property_factory("force_constants")
    symmetry = phonopy_yaml_property_factory("symmetry")
    frequency_unit_conversion_factor = phonopy_yaml_property_factory(
        "frequency_unit_conversion_factor"
    )
    version = phonopy_yaml_property_factory("version")

    phonon_supercell_matrix = phonopy_yaml_property_factory("phonon_supercell_matrix")
    phonon_dataset = phonopy_yaml_property_factory("phonon_dataset")
    phonon_supercell = phonopy_yaml_property_factory("phonon_supercell")
    phonon_primitive = phonopy_yaml_property_factory("phonon_primitive")

    def __init__(
        self, configuration=None, calculator=None, physical_units=None, settings=None
    ):
        """Init method."""
        self._data = PhelelYamlData(
            configuration=configuration,
            calculator=calculator,
            physical_units=physical_units,
        )
        self._dumper_settings = settings

    def __str__(self):
        """Return string text of yaml output."""
        pheyml_dumper = PhelelYamlDumper(
            self._data, dumper_settings=self._dumper_settings
        )
        return "\n".join(pheyml_dumper.get_yaml_lines())

    def read(self, filename):
        """Read PhelelYaml file."""
        self._data = read_phelel_yaml(
            filename,
            configuration=self._data.configuration,
            calculator=self._data.calculator,
            physical_units=self._data.physical_units,
        )
        return self

    def set_phelel_info(self, phelel: "Phelel"):
        """Store data in Phelel instance in this instance."""
        super().set_phonon_info(phelel)
        self._data.phonon_supercell_matrix = phelel.phonon_supercell_matrix
        self._data.phonon_dataset = phelel.phonon_dataset
        self._data.phonon_primitive = phelel.phonon_primitive
        self._data.phonon_supercell = phelel.phonon_supercell


def read_phelel_yaml(
    filename, configuration=None, calculator=None, physical_units=None
) -> PhelelYamlData:
    """Read phelel.yaml like file."""
    yaml_data = load_yaml(filename)
    if isinstance(yaml_data, str):
        msg = f'Could not load "{filename}" properly.'
        raise TypeError(msg)
    return load_phelel_yaml(
        yaml_data,
        configuration=configuration,
        calculator=calculator,
        physical_units=physical_units,
    )


def load_phelel_yaml(
    yaml_data, configuration=None, calculator=None, physical_units=None
) -> PhelelYamlData:
    """Return PhelelYamlData instance loading yaml data.

    Parameters
    ----------
    yaml_data : dict

    """
    pheyml_loader = PhelelYamlLoader(
        yaml_data,
        configuration=configuration,
        calculator=calculator,
        physical_units=physical_units,
    )
    pheyml_loader.parse()
    return pheyml_loader.data



"""velph command line tool / velph-nac."""

import pathlib

import click

from phelel.velph.cli import cmd_root
from phelel.velph.cli.nac.generate import write_input_files


@cmd_root.group("nac")
@click.help_option("-h", "--help")
def cmd_nac():
    """Choose nac options."""
    pass


@cmd_nac.command("generate")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str):
    """Generate relax input files."""
    if not pathlib.Path("POTCAR").exists():
        click.echo('"POTCAR" not found in current directory.')

    write_input_files(pathlib.Path(toml_filename))



"""Option parser."""


def get_parser(load_phelel_yaml=False):
    """Return ArgumentParser instance."""
    import argparse

    parser = argparse.ArgumentParser(
        description="Electron-phonon interaction calculator"
    )
    parser.add_argument(
        "--amplitude",
        dest="displacement_distance",
        type=float,
        default=None,
        help="Distance of displacements",
    )
    if not load_phelel_yaml:
        parser.add_argument(
            "-c",
            "--cell",
            dest="cell_filename",
            metavar="FILE",
            default=None,
            help="Read unit cell",
        )
    parser.add_argument(
        "--cd",
        "--create-derivatives",
        nargs="+",
        dest="create_derivatives",
        metavar="DIR",
        default=None,
        help=(
            "Calculate dVdu and dDijdu and create phelel_disp.hdf5. "
            "Directory names of perfect "
            "and displacement supercells are given as arguments."
        ),
    )
    if load_phelel_yaml:
        parser.add_argument(
            "--config",
            dest="conf_filename",
            metavar="FILE",
            default=None,
            help="Read phelel configuration file",
        )
    if not load_phelel_yaml:
        parser.add_argument(
            "-d",
            "--disp",
            dest="is_displacement",
            action="store_true",
            default=False,
            help="As first stage, get least displacements",
        )
        parser.add_argument(
            "--dim",
            nargs="+",
            dest="supercell_dimension",
            metavar="INT",
            default=None,
            help=(
                "Supercell dimensions with three integers or matrix with " "9 integers"
            ),
        )
        parser.add_argument(
            "--dim-phonon",
            nargs="+",
            dest="phonon_supercell_dimension",
            default=None,
            help=(
                "Supercell dimensions for phonon with three integers or "
                "matrix with 9 integers"
            ),
        )
    parser.add_argument(
        "--fft-mesh",
        nargs="+",
        dest="fft_mesh_numbers",
        default=None,
        help="FFT mesh numbers used in primitive cell",
    )
    parser.add_argument(
        "--finufft-eps",
        dest="finufft_eps",
        type=float,
        default=None,
        help="Accuracy of finufft interpolation (default=1e-6)",
    )
    parser.add_argument(
        "--loglevel",
        dest="log_level",
        type=int,
        metavar="INT",
        default=None,
        help="Log level",
    )
    parser.add_argument(
        "--nosym",
        dest="is_nosym",
        action="store_true",
        default=None,
        help="Symmetry is not imposed.",
    )
    parser.add_argument(
        "--nodiag",
        dest="is_nodiag",
        action="store_true",
        default=False,
        help="Set displacements parallel to axes",
    )
    parser.add_argument(
        "--pa",
        "--primitive-axis",
        "--primitive-axes",
        nargs="+",
        dest="primitive_axes",
        default=None,
        help="Same as PRIMITIVE_AXES tags",
    )
    parser.add_argument(
        "--pm",
        dest="is_plusminus_displacements",
        action="store_true",
        default=False,
        help="Set plus minus displacements",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=False,
        help="Print out smallest information",
    )
    parser.add_argument(
        "--qpoints",
        nargs="+",
        dest="qpoints",
        metavar="NUM",
        default=None,
        help="Phonon q-points for specific sampling to calculate |g|",
    )
    parser.add_argument(
        "--read-qpoints",
        dest="read_qpoints",
        action="store_true",
        default=False,
        help="Read QPOITNS file for specific phonon sampling to calculate |g|",
    )
    parser.add_argument(
        "--tolerance",
        dest="symmetry_tolerance",
        type=float,
        default=None,
        help="Symmetry tolerance to search",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Detailed run-time information is displayed.",
    )
    parser.add_argument(
        "--srfs",
        "--subtract-residual-forces",
        dest="subtract_rfs",
        action="store_true",
        default=False,
        help="Subtract residual forces for force constants calculation.",
    )
    if load_phelel_yaml:
        parser.add_argument("filename", nargs="*", help="phelel.yaml like file")
    else:
        parser.add_argument("filename", nargs="*", help="Phelel configure file")

    return parser



"""Wrapper functions to calculate derivatives."""

from __future__ import annotations

import os
import pathlib
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from phonopy import Phonopy
from phonopy.file_IO import get_born_parameters
from phonopy.interface.vasp import parse_set_of_forces

from phelel import Phelel
from phelel.api_phelel import PhelelDataset
from phelel.interface.vasp.file_IO import (
    read_inwap_vaspouth5,
    read_inwap_yaml,
    read_local_potential,
    read_local_potential_vaspouth5,
    read_PAW_Dij_qij,
    read_PAW_Dij_qij_vaspouth5,
)


def read_files(
    phelel: Phelel,
    dir_names: Sequence[str],
    phonon_dir_names: Optional[Sequence[str]] = None,
    subtract_rfs: bool = False,
    log_level: int = 0,
) -> PhelelDataset:
    """Load files needed to create derivatives."""
    inwap_path = pathlib.Path(dir_names[0]) / "inwap.yaml"
    if inwap_path.exists():
        inwap_per = read_inwap_yaml(inwap_path)
    else:
        # try reading from vaspout.h5
        inwap_path = pathlib.Path(dir_names[0]) / "vaspout.h5"
        inwap_per = read_inwap_vaspouth5(inwap_path)

    if inwap_per["nions"] != len(phelel.supercell):
        raise ValueError(
            "Number of ions in the supercell is different from the number of atoms "
            "in the inwap.yaml or vaspout.h5 file."
        )

    if log_level:
        print(f'"{inwap_path}" was read.')
    dataset, _ = _get_datasets(phelel)
    loc_pots = _read_local_potentials(dir_names, inwap_per, log_level=log_level)
    Dijs, qijs = _read_PAW_strength_and_overlap(
        dir_names, inwap_per, log_level=log_level
    )
    if phonon_dir_names is None:
        _dir_names = dir_names
    else:
        _dir_names = phonon_dir_names
    vasprun_filenames = _get_vasprun_filenames(_dir_names, log_level=log_level)
    forces = read_forces_from_vasprunxmls(
        vasprun_filenames,
        phelel.phonon,
        subtract_rfs=subtract_rfs,
        log_level=log_level,
    )

    if forces[0].shape[0] != len(phelel.phonon_supercell):
        raise ValueError(
            "Number of ions in the phonon supercell is different from the number of "
            "atoms in the vasprun.xml file."
        )

    nac_params = _read_born(phelel.phonon)
    phelel.phonon.forces = forces
    if nac_params:
        phelel.phonon.nac_params = nac_params
    return PhelelDataset(
        local_potentials=loc_pots,
        Dijs=Dijs,
        qijs=qijs,
        lm_channels=inwap_per["lm_orbitals"],
        dataset=dataset,
        forces=np.array(forces, dtype="double", order="C"),
    )


def create_derivatives(
    phelel: Phelel,
    dir_names: Sequence,
    subtract_rfs=False,
    log_level=0,
):
    """Calculate derivatives.

    Input files are read and derivatives are computed. The results are stored in
    Phelel instance.

    When the number of dir_names is equivalent to the number of displacements
    for the el-ph calculation, the same directories are used for calculating
    force constants. To calculate force constants from another directories,
    those directories have to be appdended after the directory names for
    el-ph.

    % phelel -d --dim 2 2 2 --pa auto
    % phelel --fft-mesh 18 18 18 --cd perfect disp-001

    """
    if log_level > 0:
        print("Calculation of dV/du, dDij/du, and force constants")
    dataset, phonon_dataset = _get_datasets(phelel)
    num_disp = len(dataset["first_atoms"]) + 1
    num_disp_ph = len(phonon_dataset["first_atoms"]) + 1
    if len(dir_names) == num_disp:
        phonon_dir_names = dir_names
    elif len(dir_names) == num_disp + num_disp_ph:
        phonon_dir_names = dir_names[num_disp:]
    else:
        raise RuntimeError("Number of dir_names is wrong.")

    phe_input = read_files(
        phelel,
        dir_names,
        phonon_dir_names=phonon_dir_names,
        subtract_rfs=subtract_rfs,
        log_level=log_level,
    )
    if phelel.fft_mesh is not None:
        phelel.run_derivatives(phe_input)

    # phelel.Rij = read_Rij(dir_names[0], inwap_per)


def read_Rij(dir_name, inwap_per):
    """Read Rij."""
    return read_PAW_Dij_qij(inwap_per, "%s/PAW-Rnij.bin" % dir_name, is_Rij=True)


def read_forces_from_vasprunxmls(
    vasprun_filenames: Union[list, tuple],
    phonon: Phonopy,
    subtract_rfs=False,
    log_level=0,
):
    """Read forces from vasprun.xml's and read NAC params from BORN."""
    calc_dataset = parse_set_of_forces(
        len(phonon.supercell), vasprun_filenames, verbose=False
    )
    forces = calc_dataset["forces"]

    if subtract_rfs:
        _forces = [fset - forces[0] for fset in forces[1:]]
        if log_level:
            print(
                "Residual forces of perfect supercell were subtracted from "
                "supercell forces."
            )
    else:
        _forces = forces[1:]

    return _forces


def _get_vasprun_filenames(dir_names, log_level=0):
    vasprun_filenames = []
    for dir_name in dir_names:
        filename = next(pathlib.Path(dir_name).glob("vasprun.xml*"))
        vasprun_filenames.append(filename)
        if log_level:
            print('"%s" was read.' % filename)
    return vasprun_filenames


def _read_born(phonon: Phonopy, log_level=0):
    if pathlib.Path("BORN").exists():
        with open("BORN", "r") as f:
            nac_params = get_born_parameters(
                f, phonon.primitive, phonon.primitive_symmetry
            )
            if log_level:
                print('"BORN" was read.')
        return nac_params
    else:
        return None


def _get_datasets(phelel: Phelel) -> tuple:
    """Return inwap dataset and phonopy dataset."""
    if "first_atoms" in phelel.dataset:
        dataset = phelel.dataset
        if "first_atoms" in phelel.phonon_dataset:
            phonon_dataset = phelel.phonon_dataset
        else:
            phonon_dataset = dataset
    else:
        raise RuntimeError("Displacement dataset has to be stored in Phelel instance.")

    return dataset, phonon_dataset


def _read_local_potentials(
    dir_names: Union[str, bytes, os.PathLike], inwap_per: dict, log_level: int = 0
) -> list[np.ndarray]:
    loc_pots = []
    for dir_name in dir_names:
        # Note glob returns a generator.
        possible_locpot_paths = list(
            pathlib.Path(dir_name).glob("LOCAL-POTENTIAL.bin*")
        )
        if possible_locpot_paths:
            locpot_path = possible_locpot_paths[0]
            loc_pots.append(read_local_potential(inwap_per, filename=locpot_path))
        else:
            locpot_path = pathlib.Path(dir_name) / "vaspout.h5"
            loc_pots.append(
                read_local_potential_vaspouth5(inwap_per, filename=locpot_path)
            )
        if log_level:
            print(f'"{locpot_path}" was read.')
    return loc_pots


def _read_PAW_strength_and_overlap(
    dir_names, inwap_per, log_level=0
) -> tuple[list[np.ndarray], list[np.ndarray]]:
    Dijs = []
    qijs = []
    for dir_name in dir_names:
        # Note glob returns a generator.
        possible_Dij_path = list(pathlib.Path(dir_name).glob("PAW-STRENGTH.bin*"))
        possible_qij_path = list(pathlib.Path(dir_name).glob("PAW-OVERLAP.bin*"))
        if possible_Dij_path and possible_qij_path:
            Dij_path = possible_Dij_path[0]
            qij_path = possible_qij_path[0]
            Dijs.append(read_PAW_Dij_qij(inwap_per, Dij_path))
            qijs.append(read_PAW_Dij_qij(inwap_per, qij_path))
            if log_level:
                print(f'"{Dij_path}" and "{qij_path}" were read.')
        else:
            Dij_qij_path = pathlib.Path(dir_name) / "vaspout.h5"
            dij, qij = read_PAW_Dij_qij_vaspouth5(Dij_qij_path)
            Dijs.append(dij)
            qijs.append(qij)
            if log_level:
                print(f'"{Dij_qij_path}" was read.')
    return Dijs, qijs



"""Implementation of velph-phelel-differentiate."""

import os
import pathlib
from typing import Union

import click

from phelel import Phelel
from phelel.interface.vasp.derivatives import create_derivatives
from phelel.velph.cli.utils import get_num_digits


def run_derivatives(
    phe: Phelel,
    subtract_residual_forces: bool = True,
    dir_name: Union[str, bytes, os.PathLike] = "phelel",
) -> bool:
    """Calculate derivatives and write phelel_params.hdf5."""
    dir_names = []
    nd = get_num_digits(phe.supercells_with_displacements)
    for i, _ in enumerate(
        [
            phe.supercell,
        ]
        + phe.supercells_with_displacements
    ):
        id_number = f"{i:0{nd}d}"
        filepath = pathlib.Path(f"{dir_name}/disp-{id_number}")
        if filepath.exists():
            if _check_files_exist(filepath):
                dir_names.append(filepath)
            else:
                click.echo(f'Necessary file not found in "{filepath}".', err=True)
                return False
        else:
            click.echo(f'"{filepath}" does not exist.', err=True)
            return False

    if phe.phonon_supercell_matrix is not None:
        nd = get_num_digits(phe.phonon_supercells_with_displacements)
        for i, _ in enumerate(
            [
                phe.phonon_supercell,
            ]
            + phe.phonon_supercells_with_displacements
        ):
            id_number = f"{i:0{nd}d}"
            filepath = pathlib.Path(f"{dir_name}/ph-disp-{id_number}")
            if filepath.exists():
                dir_names.append(filepath)
            else:
                click.echo(f'"{filepath}" does not exist.', err=True)
                return False

    create_derivatives(
        phe,
        dir_names,
        subtract_rfs=subtract_residual_forces,
        log_level=0,
    )

    return True


def _check_files_exist(filepath: pathlib.Path) -> bool:
    if not _check_file_exists(filepath, "vasprun.xml"):
        click.echo(f'"{filepath}/vasprun.xml" not found.', err=True)
        return False
    if _check_four_files_exist(filepath):
        return True
    else:
        if (filepath / "vaspout.h5").exists():
            click.echo(f'Found "{filepath}/vaspout.h5".', err=True)
            return True
        else:
            for filename in (
                "inwap.yaml",
                "LOCAL-POTENTIAL.bin",
                "PAW-STRENGTH.bin",
                "PAW-OVERLAP.bin",
            ):
                if not _check_file_exists(filepath, filename):
                    click.echo(f'"{filepath}/{filename}" not found.', err=True)
            return False


def _check_four_files_exist(filepath: pathlib.Path) -> bool:
    for filename in (
        "inwap.yaml",
        "LOCAL-POTENTIAL.bin",
        "PAW-STRENGTH.bin",
        "PAW-OVERLAP.bin",
    ):
        if not _check_file_exists(filepath, filename):
            return False
    return True


def _check_file_exists(filepath: pathlib.Path, filename: str) -> bool:
    """Check if the necessary file exists.

    The file can be compressed with xz, etc.

    The file names that can be checked are:
        inwap.yaml
        LOCAL-POTENTIAL.bin
        PAW-STRENGTH.bin
        PAW-OVERLAP.bin
        vasprun.xml

    """
    return bool(list(pathlib.Path(filepath).glob(f"{filename}*")))



"""Utilities to handle scheduler scripts."""

from __future__ import annotations

import copy
from typing import Optional, Union


def get_sge_scheduler_script(
    toml_scheduler_dict: dict,
    job_id: Optional[Union[str, int]] = None,
) -> str:
    """Return scheduler script of SGE.

    This is called when scheduler_name = "sge".

    Supported parameters
    --------------------
    Necessary tags:
        job_name
        mpirun_command
        vasp_binary
        prepend_text
        append_text
    Optional tags:
        pe
        walltime : [hours:minutes:]seconds, e.g., 12:00:00

    """
    header = """#!/bin/bash
#$ -cwd
#$ -S /bin/bash
#$ -m n
#$ -N {job_name}
#$ -V
#$ -o _scheduler-stdout.txt
#$ -e _scheduler-stderr.txt
"""
    params = []
    if "pe" in toml_scheduler_dict:
        params.append("#$ -pe {pe}")
    if "walltime" in toml_scheduler_dict:
        params.append("#$ -l h_rt={walltime}")
    if params:
        params_header = "\n".join(params) + "\n"
    else:
        params_header = ""

    commands = """
{prepend_text}

{mpirun_command} {vasp_binary}

{append_text}
"""
    scheduler_template = header + params_header + commands

    return get_custom_schedular_script(scheduler_template, toml_scheduler_dict, job_id)


def get_slurm_scheduler_script(
    toml_scheduler_dict: dict,
    job_id: Optional[Union[str, int]] = None,
) -> str:
    """Return scheduler script of SLURM.

    This is called when scheduler_name = "slurm".

    Supported parameters
    --------------------
    Necessary tags:
        job_name
        mpirun_command
        vasp_binary
        prepend_text
        append_text
        partition
        nodes
        ntasks

    """
    scheduler_template = """#!/bin/bash
#SBATCH --job-name={job_name}
#SBATCH --partition={partition}
#SBATCH --nodes={nodes}
#SBATCH --ntasks={ntasks}
#SBATCH --time=96:00:00      # Time limit hrs:min:sec
#SBATCH --output=ci_%j.log   # Standard output and error log

{prepend_text}

{mpirun_command} {vasp_binary}

{append_text}
"""

    return get_custom_schedular_script(scheduler_template, toml_scheduler_dict, job_id)


def get_custom_schedular_script(
    template: str, toml_scheduler_dict: dict, job_id: Optional[Union[str, int]]
) -> str:
    """Return scheduler script with given template.

    This is called when scheduler_name = "custom".

    """
    sched_dict = copy.deepcopy(toml_scheduler_dict)
    if job_id is not None:
        sched_dict["job_name"] += f"-{job_id}"
    return template.format(**sched_dict)



"""Routines to handle spherical harmonics rotation.

Old implementation`SHRotationMatricesEularAngle` relies on `EulerAngles`,
`LyMatrices`, `WignerMatricesEularAngle`. These are currently unused.

The current implementation `SHRotationMatrice`s depends on `LxLyLzMatrices` and
get_n_and_rotation_order.

Both implementations call `make_C`.

Agreement of these two implementations are checked in the test. The new
implementation is expected to be numerically more robust.

"""

from __future__ import annotations

from typing import Optional

import numpy as np
from phonopy.structure.cells import determinant


class EulerAngles:
    """Calculate Euler angles.

    Implements Z1Y2Z3 given in Euler angles page of wikipedia.

    """

    def __init__(self, R: np.ndarray):
        """Init method.

        Proper rotation of R is transformed to Eular angles.

        Parameters
        ----------
        R(np_double): Rotation matrix in Cartesian coordinates

        """
        self.R = np.array(R, dtype="double", order="C")
        self._angles = None

    @property
    def angles(self) -> Optional[np.ndarray]:
        """Return Eular angles."""
        return self._angles

    def run(self):
        """Calculate Euler angles.

        R = R_z(alpha) * R_y(beta) * R_z(gamma)

        cb: cos(beta)
        sb: sin(beta) = sqrt(1 - cos(beta)**2)

        ca: cos(alpha)
        sa: sin(alpha)
        cg: cos(gamma)
        sg: sin(gamma)

        """
        R = self.R * np.linalg.det(self.R)
        cb = R[2, 2]

        cb2 = cb**2
        if cb2 > 1:
            if cb2 < 1 + 1e-10:
                cb2 = 1
            else:
                raise RuntimeError("Rotation matrix is largely distorted.")
        # sqrt may not be very numerically accurate if cb**2 is close to 1.
        sb = np.sqrt(1 - cb2)

        if 1 - cb < 1e-10:  # beta ~ 0
            # A choice with gamma=0
            cg = 1
            sg = 0
            ca = R[0, 0]
            sa = R[1, 0]
        elif cb + 1 < 1e-10:  # beta ~ pi
            # A choice with gamma=0
            cg = 1
            sg = 0
            ca = -R[0, 0]
            sa = -R[1, 0]
        else:
            ca = R[0, 2] / sb
            sa = R[1, 2] / sb
            cg = -R[2, 0] / sb
            sg = R[2, 1] / sb

        self._angles = np.array([ca, sa, cb, sb, cg, sg], dtype="double")

    def recover_R(self) -> np.ndarray:
        """Recover R matrix from angles.

        This is mainly used for the test.

        """
        R = np.zeros((3, 3), dtype="double", order="C")
        ca, sa, cb, sb, cg, sg = self.angles
        R[0, 0] = ca * cb * cg - sa * sg
        R[0, 1] = -ca * cb * sg - sa * cg
        R[0, 2] = ca * sb
        R[1, 0] = sa * cb * cg + ca * sg
        R[1, 1] = -sa * cb * sg + ca * cg
        R[1, 2] = sa * sb
        R[2, 0] = -sb * cg
        R[2, 1] = sb * sg
        R[2, 2] = cb

        return R * np.linalg.det(self.R)


class LyMatrices:
    """Create matrix representation of <lm'|Ly|lm>.

    ihbar/2 is not multiplied.

    """

    def __init__(self, l_max: int = 3):
        """Init method."""
        self._l_max = l_max
        self._Ly = None

    @property
    def Ly(self) -> Optional[list[np.ndarray]]:
        """Return <lm'|Ly|lm> for all l channel."""
        return self._Ly

    def run(self):
        """Calculate <lm'|Ly|lm> for all l channel."""
        self._Ly = []
        for ll in range(self._l_max + 1):
            self._Ly.append(self._get_Ly(ll))

    def _get_Ly(self, ll: int) -> np.ndarray:
        """Compute <lm'|Ly|lm> at l.

        i hbar/2 is not multiplied.

        """
        Ly = np.zeros((2 * ll + 1, 2 * ll + 1), dtype="double", order="C")
        for mp in range(-ll, ll + 1):
            for m in range(-ll, ll + 1):
                if mp == m - 1:
                    Ly[mp + ll, m + ll] += np.sqrt(ll * (ll + 1) - m * (m - 1))
                if mp == m + 1:
                    Ly[mp + ll, m + ll] -= np.sqrt(ll * (ll + 1) - m * (m + 1))
        return Ly


class WignerMatricesEularAngle:
    """Wigner matrices dm'm^l = <lm'|e^ibLy|lm>  is calcualted."""

    def __init__(self, l_max: int = 6):
        """Init method."""
        self._ly = LyMatrices(l_max=l_max)
        self._ly.run()
        self._d = None

    @property
    def d(self) -> Optional[list[np.ndarray]]:
        """Return <lm'|e^ibLy|lm> for all l channels."""
        return self._d

    def run(self, cb: float, sb: float):
        """Calculate <lm'|e^ibLy|lm> for all l channels."""
        self._d = []
        for Ly_l in self._ly.Ly:
            w, U = np.linalg.eigh(0.5j * Ly_l)  # i/2 is multipled to Ly_l.
            exp_w = (cb - 1j * sb) ** w
            self._d.append(np.dot(U * exp_w, U.T.conj()))


class SHRotationMatricesEularAngle:
    """Rotation matrix for spherical harmonics, Dm'm^l.

    Dm'm = e^-iam' dm'm(b) e^-igm (a: alpha, b: beta, g: gamma)
    for usual spherical harmonics and
    Delta_m'm for real spherical harmonics

    """

    def __init__(self, R: np.ndarray, lattice: np.ndarray):
        """Init method.

        Parameters
        ----------
        R: ndarray
            Rotation matrix wrt basis vectors (not Cartesian).
            shape=(3, 3).
        lattice: ndarray.
            Basis vectors in column vectors.
            shape=(3, 3)

        """
        self._R_cart = lattice @ R @ np.linalg.inv(lattice)
        self._detR = np.linalg.det(R)
        self._Delta = None

    @property
    def Delta(self) -> list[np.ndarray]:
        """Return rotation matrix of spherical harmonics."""
        return self._Delta

    def run(self, l_max: int = 6):
        """Rotate spherical harmonics matrices."""
        prop_R_cart = self._R_cart * self._detR
        ea = EulerAngles(prop_R_cart)
        ea.run()
        ca, sa, cb, sb, cg, sg = ea.angles
        wm = WignerMatricesEularAngle(l_max=l_max)
        wm.run(cb, sb)
        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        delta = []
        for ll, d_l in enumerate(wm.d):
            D = np.array(d_l, dtype=dtype, order="C")
            for mp in range(-ll, ll + 1):
                D[mp + ll, :] *= (ca - 1j * sa) ** mp
            for m in range(-ll, ll + 1):
                D[:, m + ll] *= (cg - 1j * sg) ** m
            D *= self._detR**ll
            C = make_C(ll, D)
            delta.append(C.conj() @ D @ C.T)
        self._Delta = delta


class LxLyLzMatrices:
    """Create matrix representation of <lm'|Lx|lm>, <lm'|Ly|lm>, <lm'|Lz|lm>.

    hbar is not multiplied. Only integer l's are considered.

    """

    def __init__(self, l_max: int = 6):
        """Init method."""
        self._l_max = l_max
        self._Lx = None
        self._Ly = None
        self._Lz = None

    @property
    def Lx(self) -> Optional[list[np.ndarray]]:
        """Return <lm'|Lx|lm> for all l channel."""
        return self._Lx

    @property
    def Ly(self) -> Optional[list[np.ndarray]]:
        """Return <lm'|Ly|lm> for all l channel."""
        return self._Ly

    @property
    def Lz(self) -> Optional[list[np.ndarray]]:
        """Return <lm'|Lz|lm> for all l channel."""
        return self._Lz

    def run(self):
        """Calculate <lm'|Lx|lm>, <lm'|Lz|lm>, <lm'|Lz|lm> for all l channel."""
        self._Lx = []
        self._Ly = []
        self._Lz = []
        for j in range(self._l_max + 1):
            self._Lx.append(self._get_Lx(j))
            self._Ly.append(self._get_Ly(j))
            self._Lz.append(self._get_Lz(j))
        return self

    def _get_Lx(self, ll: int) -> np.ndarray:
        lx = np.zeros((2 * ll + 1, 2 * ll + 1), dtype=complex, order="C")
        for mp in range(-ll, ll + 1):
            for m in range(-ll, ll + 1):
                if mp == m + 1:
                    lx[mp + ll, m + ll] += np.sqrt((ll - m) * (ll + m + 1))
                if mp == m - 1:
                    lx[mp + ll, m + ll] += np.sqrt((ll + m) * (ll - m + 1))
        return lx / 2

    def _get_Ly(self, ll: int) -> np.ndarray:
        ly = np.zeros((2 * ll + 1, 2 * ll + 1), dtype=complex, order="C")
        for mp in range(-ll, ll + 1):
            for m in range(-ll, ll + 1):
                if mp == m + 1:
                    ly[mp + ll, m + ll] += -np.sqrt((ll - m) * (ll + m + 1))
                if mp == m - 1:
                    ly[mp + ll, m + ll] += np.sqrt((ll + m) * (ll - m + 1))
        return 1j * ly / 2

    def _get_Lz(self, ll: int) -> np.ndarray:
        lz = np.zeros((2 * ll + 1, 2 * ll + 1), dtype=complex, order="C")
        for mp in range(-ll, ll + 1):
            for m in range(-ll, ll + 1):
                if mp == m:
                    lz[mp + ll, m + ll] = m
        return lz


class SHRotationMatrices:
    r"""Rotation matrix for spherical harmonics, Dlm'lm(n,alpha).

    D^{lm'lm}(\mathbf{n}, \alpha) =
        <lm'|e^{-i/\hbar \mathbf{L}\cdot\mathbf{n}\alpha}|lm>.

    """

    def __init__(self, R: np.ndarray, lattice: np.ndarray, lxlylz: LxLyLzMatrices):
        """Init method.

        Parameters
        ----------
        R : ndarray
            Rotation matrix wrt basis vectors (not Cartesian).
            shape=(3, 3).
        lattice : ndarray.
            Basis vectors in column vectors.
            shape=(3, 3)
        lxlylz : LxLyLzMatrices

        """
        self._lxlylz = lxlylz
        self._d = None
        self._Delta = None
        self._n_vec, self._r_order, self._detR = get_n_and_rotation_order(R, lattice)
        if self._r_order == 1:
            self._alpha = 0.0
        else:
            self._alpha = 2 * np.pi / self._r_order

    @property
    def Delta(self) -> Optional[list[np.ndarray]]:
        """Return rotation matrix of spherical harmonics."""
        return self._Delta

    def run(self):
        """Rotate spherical harmonics matrices."""
        self._compute_d()
        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        delta = []
        for ll, d_l in enumerate(self._d):
            D = np.array(d_l, dtype=dtype, order="C")
            D *= self._detR**ll
            C = make_C(ll, D)
            delta.append(C.conj() @ D @ C.T)
        self._Delta = delta

    def _compute_d(self):
        r"""Calculate <lm'|e^iL.n\alpha|lm> for all l channels."""
        self._d = []
        n = self._n_vec
        for lx, ly, lz in zip(self._lxlylz.Lx, self._lxlylz.Ly, self._lxlylz.Lz):
            ln = n[0] * lx + n[1] * ly + n[2] * lz
            w, U = np.linalg.eigh(ln)
            self._d.append((U * np.exp(-1j * w * self._alpha)) @ U.T.conj())


def get_n_and_rotation_order(
    R: np.ndarray, L: np.ndarray
) -> tuple[np.ndarray, int, int]:
    """Return proper-rotation axis, order of rotation, and det(R).

    r_order = 1 :
        n = (0, 0, 1), alpha = 0.
    r_order = 2 :
        n is taken as one of eigvecs of proper rotation in Cartesian
        coordinates whose eigenvalue = -1. alpha = pi.
    r_order in (3, 4, 6) :
        Given e_1=(1, 0, 0), e_2=(0, 1, 0), e_3=(0, 0, 1),
            v = R.e - e
        is perpendicular to the rotationa axis. Compute cross product,
            v_a = cross(v, R.v).
        The unit vector of the rotation axis is
            n = v_a / |v_a|
        Order of rotation (N) is obtained by iterating rotation matrix until
        being identity. The rotation angle is
            alpha = 2pi / N.

    Parameters
    ----------
    R: ndarray
        Rotation matrix wrt basis vectors (not Cartesian). shape=(3, 3).
    L: ndarray.
        Basis vectors in column vectors. shape=(3, 3)

    Returns
    -------
    tuple :
        n : Unit vector of rotation axis
        r_order : Order of rotation (0 <= theta < pi).
        detR : Determinant of rotation matrix, expected -1 or 1.

    """
    detR = determinant(R)
    R_prop = R * detR
    identity = np.eye(3, dtype=int)
    if np.array_equal(R_prop, identity):
        return np.array([0, 0, 1], dtype="double"), 1, detR

    rbyr = identity.copy()
    for r_order in range(1, 7):  # noqa B007
        rbyr = rbyr @ R_prop
        if np.array_equal(rbyr, identity):
            break

    assert r_order > 1

    r_cart = L @ R @ np.linalg.inv(L) * detR

    if r_order == 2:
        eigvals, eigvecs = np.linalg.eig(r_cart)
        n = None
        for v, vec in zip(eigvals, eigvecs.T):
            if abs(v - 1) < 1e-5:
                n = vec

        assert n is not None

        return n, 2, detR
    else:
        assert r_order in (3, 4, 6)
        vecs = r_cart - identity
        vec = vecs[np.argmax(np.linalg.norm(vecs, axis=0))]
        n = np.cross(vec, r_cart @ vec)
        n /= np.linalg.norm(n)
        return n, r_order, detR


def make_C(ll: int, D: np.ndarray) -> np.ndarray:
    """Make C matrix.

    See L. Chaput et al., Phys. Rev. B 100, 174304 (2019).

    """
    C = np.zeros_like(D)
    C[ll, ll] = np.sqrt(2)
    for i in range(ll):
        C[i, i] = 1j
        C[i, 2 * ll - i] = -1j * (-1) ** (ll - i)
        C[2 * ll - i, i] = 1
        C[2 * ll - i, 2 * ll - i] = (-1) ** (ll - i)
    C /= np.sqrt(2)

    assert (abs(C.conj().T - np.linalg.inv(C)) < 1e-5).all()

    return C



"""Test for Phelel class."""

import pathlib

import h5py
import numpy as np

from phelel import Phelel
from phelel.file_IO import read_phelel_params_hdf5
from phelel.utils.data import cmplx2real

cwd = pathlib.Path(__file__).parent


def test_api_phelel_C111(phelel_C111: Phelel):
    """Test by diamond conv. unit cell 1x1x1."""
    filename = cwd / "phelel_params_C111.hdf5"
    _compare(filename, phelel_C111)


def test_api_phelel_NaCl111(phelel_NaCl111: Phelel):
    """Test by NaCl conv. unit cell 1x1x1."""
    filename = cwd / "phelel_params_NaCl111.hdf5"
    _compare(filename, phelel_NaCl111)


def test_api_phelel_CdAs2_111(phelel_CdAs2_111: Phelel):
    """Test by CdAs2 conv. unit cell 1x1x1 (I-centred tetragonal)."""
    filename = cwd / "phelel_params_CdAs2_111.hdf5"
    _compare(filename, phelel_CdAs2_111)


def test_read_phelel_params_hdf5(phelel_CdAs2_111: Phelel):
    """Test reading phelel_params using CdAs2."""
    filename = cwd / "phelel_params_CdAs2_111.hdf5"
    dVdu, dDijdu, _, _ = read_phelel_params_hdf5(filename=filename)
    phe_ref = phelel_CdAs2_111

    np.testing.assert_allclose(dVdu.dVdu, phe_ref.dVdu.dVdu, rtol=1e-5, atol=1e-5)
    np.testing.assert_allclose(
        dDijdu.dDijdu, cmplx2real(phe_ref.dDijdu.dDijdu), rtol=1e-5, atol=1e-5
    )


def _compare(filename: pathlib.Path, phe: Phelel):
    """Assert results.

    shortest_vectors and shortest_vector_multiplicities are included at later
    versions of Phelel. So, these are not included in old reference files.

    """
    with h5py.File(filename, "r") as f:
        dVdu_ref = f["dVdu"][:]
        dDijdu_ref = f["dDijdu"][:]

        dVdu = cmplx2real(phe.dVdu.dVdu)
        dDijdu = cmplx2real(phe.dDijdu.dDijdu)
        np.testing.assert_allclose(dVdu, dVdu_ref, rtol=1e-5, atol=1e-5)
        np.testing.assert_allclose(dDijdu, dDijdu_ref, rtol=1e-5, atol=1e-5)

        if "shortest_vectors" in f:
            shortest_vectors_ref = f["shortest_vectors"][:]
            multiplicities_ref = f["shortest_vector_multiplicities"][:]

            shortest_vectors, multiplicities = phe.primitive.get_smallest_vectors()
            np.testing.assert_array_equal(
                shortest_vectors.shape, shortest_vectors_ref.shape
            )
            np.testing.assert_array_equal(multiplicities, multiplicities_ref)



"""Calculate derivative of local potential."""

from __future__ import annotations

import io
import os
import textwrap
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional, Union

if TYPE_CHECKING:
    from phelel.base.Dij_qij import DeltaDijQij

import numpy as np
from phonopy.interface.vasp import get_vasp_structure_lines
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import SNF3x3, determinant, get_supercell
from phonopy.structure.symmetry import Symmetry
from phonopy.utils import similarity_transformation

from phelel.interface.vasp.file_IO import get_CHGCAR
from phelel.utils.data import real2cmplx
from phelel.utils.lattice_points import get_lattice_points
from phelel.utils.spinor import SpinorRotationMatrices


class DeltaLocalPotential:
    """Container to store change in local potential by atomic displacement.

    Note
    ----
    ncdij = 1 without spin
    ncdij = 2 with collinear spin (up and down)
    ncdij = 4 with non-collinear spin (2x2 localpotential)

    Attributes
    ----------
    dV : ndarray
        Difference of local potentials with and without an atom displacement
        dtype='complex128'
        shape=(ncdij, nz, ny, nx)
    displacements : dict
        See docstring of __init__.

    """

    def __init__(
        self, V_loc_per: np.ndarray, V_loc_disp: np.ndarray, displacement: dict
    ):
        """Init method.

        Parameters
        ----------
        V_loc_per : ndarray
            Local potential of perfect supercell
            dtype='complex128'
            shape=(ncdij, nz, ny, nx)
        V_loc_disp : ndarray
            Local potential of sueprcell with a displacement
            dtype=''complex128''
            shape=(ncdij, nz, ny, nx)
        displacement : dict
            Displacement of one atom
            keys :
                'displacement' : Displacement in Cartesian coordinates
                'number' : Index of displaced atom

        """
        self.dV = V_loc_disp - V_loc_per
        self.displacement = displacement

    def write(
        self,
        supercell: PhonopyAtoms,
        filename: Optional[Union[str, bytes, os.PathLike, io.IOBase]] = None,
        verbose: bool = False,
    ):
        """Write local potential to a text file."""
        if filename:
            _filename = filename
        else:
            _filename = "locpot.dat"

        if verbose:
            print(f'dV_loc is written in "{_filename}" in LOCPOT like ' "format.")

        header = "\n".join(get_vasp_structure_lines(supercell))
        locpot = get_CHGCAR(self.dV[0].real, header)  # only cdij=0

        with open(_filename, "w") as w:
            w.write(locpot)


class LocalPotentialInterpolationNUFFT:
    """Interpolation of local potential by non-uniform FFT.

    For one displaced atom and maybe also symmetrically equivalent atoms to the
    dispalced atom.

    There are two loops:
    1) Atomic diplacement directions (outer)
       iFFT of dV on supercell FFT mesh
    2) Symmetry operations (inner)
       NUFFT of iFFT(dV) onto primitive cell FFT mesh grid where
       the grid is rotated by symmetry operations passively.

    Attributes
    ----------
    p2s_matrix : ndarray
        Transformation matrix to supercell from primitive cell
        dtype='int_'
        shape=(3,3)
    grid_points : ndarray
        Points to be used to compute <psi|dV_loc|psi'> given as
        coordinates of supercell. Coordinates are given as
        [[x1, y1, z1],
         [x2, y2, z2],
         ...         ]
        dtype='double'
        shape=(prod(fft_mesh)*det(p2s_matrix), 3)
    lattice_points : ndarray
        Lattice points in supercell in primitive cell coordinates
        dtype='int_'
        shape=(det(supercell_matrxi), 3)
    dVdu : ndarray
        Displacement derivative of local potential in supercell interpolated on
        mesh grid of primitve cell
        dtype='complex128'
        shape=(ncdij, atom_indices, 3, grid_points)
    atom_indices_returned : ndarray
        Atom indices in supercell where dV will be computed. If those indices
        don't belong to symmetrically equivalent atoms to the dispalced atom.
        dtype='int_', shape=(atom_indices,)
    dVs : list of DeltaLocalPotential
        This is given from DLocalPotential.run().

    """

    def __init__(
        self,
        fft_mesh: Union[int, float, Sequence, np.ndarray],
        p2s_matrix: np.ndarray,
        supercell: PhonopyAtoms,
        symmetry: Symmetry,
        atom_indices: Optional[np.ndarray] = None,
        nufft: Optional[str] = None,
        finufft_eps: Optional[float] = None,
        verbose: bool = True,
    ):
        """Init method.

        Parameters
        ----------
        fft_mesh : array_like
            Mesh numbers for primitive cell
            dtype='int_'
            shape=(3,)
        p2s_matrix : array_like
            Supercell matrix relative to primitve cell
            dtype='int_'
            shape=(3,3)
        supercell : PhonopyAtoms
            Perfect supercell
        symmetry : Symmetry
            Symmetry of supercell
        atom_indices : list of int, optional
            Atom indices in supercell where dV will be expected to be computed.
            If None, supposed to be all atoms. Internally only symmetrically
            equivalent atoms to the dispalced atom are selected to compute.
        nufft : str or None, optional
            'finufft' only. Default is None, which corresponds to 'finufft'.
        finufft_eps : float or None, optional
            Accuracy of finufft interpolation. Default is None, which
            corresponds to 1e-6.
        verbose : bool, optional
            To display log or not

        """
        ##########
        # Inputs #
        ##########
        self._fft_mesh = np.array(fft_mesh, dtype="int_")
        self._verbose = verbose
        self._supercell = supercell
        self._symmetry = symmetry
        self._atom_indices_in = atom_indices
        if nufft is None:
            self._nufft = "finufft"
        else:
            self._nufft = nufft
        if finufft_eps is None:
            self._finufft_eps = 1e-6
        else:
            self._finufft_eps = finufft_eps

        ##########
        # Public #
        ##########
        self._p2s_matrix = np.array(p2s_matrix, dtype="int_", order="C")
        (self._grid_points, self._lattice_points) = get_grid_points(
            fft_mesh, self._p2s_matrix
        )
        assert 3 == self._grid_points.shape[1]

        self._dVdu = None
        self._atom_indices_returned = None

        ###########
        # Private #
        ###########
        # Delta local potentials before interpolation
        self._delta_Vs: Optional[list[DeltaLocalPotential]] = None

        # counter of equivalent atoms calculations
        self._i_atom = 0
        # Sets of symmetry operations to send
        # displaced atom to ones of
        # symmetrically equivalent atoms.
        # The set is the same as
        # site-symmetry when displaced atom
        # and the one is identical.
        self._sitesym_sets = None

        # Used in the first loop for each one of symmetrically equivalent atoms
        self._dV_itpl = None
        # Interpolated dV for different displacements
        # [[dV(0,0,0), dV(0,0,1), dV(0,0,2), ..., dV(0,0,N_g)],
        #  [dV(0,1,0), dV(0,1,1), dV(0,1,2), ..., dV(0,1,N_g)],
        #  ...
        #  [dV(1,0,0), dV(1,0,1), dV(1,0,2), ..., dV(1,0,N_g)],
        #  [dV(1,1,0), dV(1,1,1), dV(1,1,2), ..., dV(1,1,N_g)],
        #  ...
        #  [dV(n,N,0), dV(n,N,1), dV(n,N,2), ..., dV(n,N,N_g)]]
        # n:   number of displacements,
        # N:   number of symmetry operations
        # N_g: number of interpolated grid points

        # Inverse Fourier transformed original dV
        self._dV_iFT = None

        # [[d_x(0,0), d_y(0,0), d_z(0,0)],
        #  [d_x(0,1), d_y(0,1), d_z(0,1)],
        #  ...
        #  [d_x(1,0), d_y(1,0), d_z(1,0)],
        #  [d_x(1,1), d_y(1,1), d_z(1,1)],
        #  ...
        #  [d_x(n,N), d_y(n,N), d_z(n,N)]]
        # n: number of displacements,
        # N: number of symmetry operations
        self._disps = None

        self._finufft_plan = None

    def __iter__(self):
        """Enable iterator."""
        return self

    def __next__(self):
        """Enable iterator."""
        if len(self._atom_indices_returned) == self._i_atom:
            self._dV_itpl = None
            self._disps = None
            raise StopIteration

        self._run_at_atom()
        self._i_atom += 1

    def next(self):
        """Enable iterator."""
        return self.__next__()

    @property
    def p2s_matrix(self):
        """Return supercell matrix."""
        return self._p2s_matrix

    @property
    def grid_points(self):
        """Return grid points."""
        return self._grid_points

    @property
    def lattice_points(self):
        """Return lattice points."""
        return self._lattice_points

    @property
    def dVdu(self):
        """Return dVdu."""
        return self._dVdu

    @property
    def atom_indices_returned(self):
        """Return atom indices."""
        return self._atom_indices_returned

    @property
    def delta_Vs(self) -> Optional[list[DeltaLocalPotential]]:
        """Return dVs."""
        return self._delta_Vs

    @delta_Vs.setter
    def delta_Vs(self, delta_Vs: list[DeltaLocalPotential]):
        """Prepare dV interpolation calculation.

        Note
        ----
        Atom index list (atom_indices_returned) is created as a union of
        atom_indices_in and symmetrically equivalent atoms of the displaced atom.

        Parameters
        ----------
        dVs : list of DeltaLocalPotentials
            Changes in local potentials by atomic displacements. This
            can be set later.

        """
        self._delta_Vs = delta_Vs
        self._i_atom = 0
        disp_atom = self._delta_Vs[0].displacement["number"]
        sitesym_sets, equiv_atoms = collect_site_symmetry_operations(
            disp_atom, self._symmetry
        )

        if self._atom_indices_in is None:
            atoms = np.arange(len(self._supercell))
        else:
            atoms = self._atom_indices_in
        self._atom_indices_returned = np.array(
            [i for i in atoms if i in equiv_atoms], dtype="int_"
        )
        sitesym_selected_indices = [
            i
            for i, eq_atom in enumerate(equiv_atoms)
            if eq_atom in self._atom_indices_returned
        ]
        self._sitesym_sets = sitesym_sets[sitesym_selected_indices]

        dtype = f'c{np.dtype("double").itemsize * 2}'
        ncdij = self._delta_Vs[0].dV.shape[0]
        self._dVdu = np.zeros(
            (ncdij, len(self._atom_indices_returned), 3, len(self._grid_points)),
            dtype=dtype,
            order="C",
        )

    def delete_dVdu(self):
        """Delete large object dVdu."""
        self._dVdu = None

    def run(self):
        """Possibly iterate over selected symmetrically equivalent atoms."""
        for _ in self:
            pass

    def _run_at_atom(self):
        """Calculate at an atom in selected symmetrically equivalent atoms.

        Called by iterator method __next__.

        np.dot(disps_inv, full_dV_itpl) is stored in self.dVdu. But full_dV_tpl
        requires large memory space, it is divided into those at site-symmetry
        operations and multiplied with disps_inv piece by piece.

        disps:
            Stacked atomic displacements. Site-symmetry runs fastest.
            shape = (n_sitesyms * n_disps, 3)
        disps_inv:
            shape = (3, n_sitesyms * n_disps)
        dVs_*:
            shape = (ncdij, n_grid_points)
        self._dVdu:
            shape = (ncdij, natom, 3, n_grid_points)

        ``dVs_iFT``, ``dVs_rotated``, and ``self._finufft_plan`` are the list of
        size ``ncdij``, where ``ncdij`` is 1 (non-magnetic), 2 (collinear
        magnetic), or 4 (non-collinear).

        Spinor rotation
        ---------------
        When non-collinear case, dV is considered as 2x2 matrix of scalar fields
        of delta V in real space with respect to each atomic displacement. This
        dV is rotated to expand symemtrized information to non-symmetrized
        information by leveraging site symmetry information for computing the
        derivative.

        """
        sitesyms = self._sitesym_sets[self._i_atom]
        rotations = self._symmetry.symmetry_operations["rotations"][sitesyms]
        translations = self._symmetry.symmetry_operations["translations"][sitesyms]
        lattice = self._supercell.cell.T
        disps = get_displacements_with_rotations(rotations, lattice, self._delta_Vs)
        disps_inv = np.linalg.pinv(disps)

        if self._verbose:
            print("Running finufft (eps=%.3e)..." % self._finufft_eps)

        ncdij = self._dVdu.shape[0]
        self._init_finufft(ncdij)

        count = 0
        for delta_V in self._delta_Vs:
            dVs_iFT = [self._get_iFFT_of_dV(dV) for dV in delta_V.dV]
            for r, t in zip(rotations, translations):
                dVs_rotated = self._rotate_dV(dVs_iFT, r, t)
                if ncdij == 4:  # Need to rotate in spin space, too.
                    dVs_spinor_rotated = rotate_delta_vals_in_spin_space(
                        dVs_rotated, r, lattice
                    )
                    dVs_rotated.clear()
                    dVs_rotated = dVs_spinor_rotated
                for i_cdij in range(ncdij):
                    self._dVdu[i_cdij, self._i_atom] += np.outer(
                        disps_inv[:, count], dVs_rotated[i_cdij]
                    )
                count += 1
                dVs_rotated.clear()
            dVs_iFT.clear()

        self._finalize_finufft()

    def _rotate_dV(
        self, dV_iFTs: list[np.ndarray], r: np.ndarray, t: np.ndarray
    ) -> list[np.ndarray]:
        """Rotate dV by rotating coordinates of delta potential passively.

        Instead of rotating delta potential, grid points are rotated.

        """
        r_inv = np.linalg.inv(r)
        t_inv = -r_inv @ t
        grid_points = self._grid_points @ r_inv.T + t_inv
        grid_points -= np.rint(grid_points)
        return self._run_finufft(grid_points, dV_iFTs)

    def _get_iFFT_of_dV(self, dV: np.ndarray) -> np.ndarray:
        """Inverse FFT."""
        dims = dV.shape
        assert 3 == len(dims)
        dV_iFT = np.fft.fftshift(np.fft.ifftn(dV))
        return np.array(dV_iFT, dtype=dV_iFT.dtype, order="C")

    def _run_finufft(
        self, grid_points: np.ndarray, dV_iFTs: list[np.ndarray]
    ) -> list[np.ndarray]:
        """Transform from uniform to non-uniform points.

        3D Type-2 transform.

            finufft.nufft3d2(z, y, x, dV_iFT, eps=self._finufft_eps)

        dV_iFT is FFT of dV whwere dV values are stored in Fortran order.
        So x, y, z are alined as (z, y, x) in nufft3d2.

        """
        retval = []
        for i, dV_iFT in enumerate(dV_iFTs):
            x, y, z = [
                np.array(v, dtype=grid_points.dtype, order="C")
                for v in (grid_points * (np.pi * 2)).T
            ]
            self._finufft_plan[i].setpts(z, y, x)
            retval.append(self._finufft_plan[i].execute(dV_iFT))
        return retval

    def _init_finufft(self, ncdij: int):
        import finufft

        dtype = f'c{np.dtype("double").itemsize * 2}'
        self._finufft_plan = [
            finufft.Plan(
                2, self._delta_Vs[0].dV[i].shape, eps=self._finufft_eps, dtype=dtype
            )
            for i in range(ncdij)
        ]

    def _finalize_finufft(self):
        self._finufft_plan.clear()


class DLocalPotential:
    """Compute derivative of local potential wrt. atomic displacements.

    dV/du.

    Note
    ----
    ncdij = 1 without spin
    ncdij = 2 with collinear spin (up and down)
    ncdij = 4 with non-collinear spin (2x2 localpotential)

    Attributes
    ----------
    p2s_matrix : ndarray
        Transformation matrix to supercell from primitive cell.
        dtype='int_', shape=(3,3)
    grid_points : ndarray
        Points to be used to compute <psi|dV_loc|psi'> given as
        coordinates of supercell. Coordinates are given as
        [[x1, y1, z1],
         [x2, y2, z2],
         ...         ]
        dtype='double'
        shape=(prod(fft_mesh)*det(p2s_matrix), 3)
    lattice_points : ndarray
        Lattice points in supercell in primitive cell coordinates.
        dtype='int_', shape=(det(supercell_matrxi), 3)
    supercell : PhonopyAtoms
        Supercell.
    symmetry : Symmetry
        Symmetry of supercell.
    dVdu : ndarray
        Displacement derivative of local potential in supercell interpolated on
        mesh grid of primitve cell.
        dtype='complex128', shape=(ncdij, atom_indices, 3, grid_points)
    atom_indices : ndarray, optional
        Atom indices in supercell where dV is computed. This is made as
        np.unique(atom_indices given at __init__). If None, all atoms in
        supercell.
        shape=(len(atom_indices),), dtype='int_'
    fft_mesh : array_like
        Mesh numbers for primitive cell.
        dtype='int_', shape=(3,)

    """

    def __init__(
        self,
        fft_mesh: Union[int, float, Sequence, np.ndarray],
        p2s_matrix: np.ndarray,
        supercell: PhonopyAtoms,
        symmetry: Optional[Symmetry] = None,
        atom_indices: Optional[np.ndarray] = None,
        nufft: Optional[str] = None,
        finufft_eps: Optional[float] = None,
        verbose: bool = True,
    ):
        """Init method.

        Parameters
        ----------
        fft_mesh : array_like
            Mesh numbers for primitive cell.
            dtype='int_', shape=(3,)
        p2s_matrix : ndarray or 2d list
            Supercell matrix relative to primitve cell.
            dtype='int_', shape=(3,3)
        supercell : PhonopyAtoms
            Supercell.
        symmetry : Symmetry, optional
            Symmetry of supercell. If None, symmetry is searched in this class
            object.
        atom_indices : list of int, optional
            Atom indices in supercell where dV will be expected to be computed.
            If None, supposed to be all atoms. Internally only symmetrically
            equivalent atoms to the dispalced atom are selected to compute.
        nufft : str or None
            'finufft' only. Default is None, which corresponds to 'finufft'.
        finufft_eps : float or None, optional
            Accuracy of finufft interpolation. Default is None, which
            corresponds to 1e-6.
        verbose : bool
            To display log or not

        """
        self._verbose = verbose

        self._supercell = supercell
        self._fft_mesh = fft_mesh
        self._p2s_matrix = p2s_matrix
        if atom_indices is None:
            self._atom_indices = np.arange(len(self._supercell), dtype="int_")
        else:
            self._atom_indices = np.array(np.unique(atom_indices), dtype="int_")
        if symmetry is None:
            self._symmetry = Symmetry(self._supercell)
        else:
            self._symmetry = symmetry
        self._nufft = nufft
        self._finufft_eps = finufft_eps
        self._lattice_points = None
        self._grid_points = None
        self._dVdu = None  # self.dVdu is provided by @property.

    @property
    def p2s_matrix(self):
        """Return supercell matrix."""
        return self._p2s_matrix

    @property
    def supercell(self):
        """Return supercell."""
        return self._supercell

    @property
    def symmetry(self):
        """Return symmetry of supercell."""
        return self._symmetry

    @property
    def atom_indices(self):
        """Return atom indices."""
        return self._atom_indices

    @property
    def fft_mesh(self):
        """Return FFT mesh."""
        return self._fft_mesh

    @property
    def dVdu(self):
        """Return dVdu.

        See detail at attribute section of this class's docstring.

        """
        return self._dVdu

    @dVdu.setter
    def dVdu(self, dVdu: np.ndarray):
        if dVdu.dtype == "double":
            _dVdu = real2cmplx(dVdu)
        else:
            _dVdu = dVdu
        if _dVdu.shape[1:] == self._get_dVdu_shape():
            self._allocate_arrays(_dVdu.shape[0])
            self._dVdu[:] = _dVdu
        else:
            raise RuntimeError(
                "Array shape[1:] disagreement is found, %s!=%s."
                % (self._get_dVdu_shape(), _dVdu.shape[1:])
            )

    @property
    def lattice_points(self):
        """Return lattice points.

        See detail at attribute section of this class's docstring.

        """
        return self._lattice_points

    @lattice_points.setter
    def lattice_points(self, lattice_points: np.ndarray):
        multi = determinant(self.p2s_matrix)
        if lattice_points.shape == (multi, 3):
            self._lattice_points = np.array(lattice_points, dtype="int_", order="C")
        else:
            raise RuntimeError(
                "Array shape disagreement is found, %s!=%s."
                % ((multi, 3), lattice_points.shape)
            )

    @property
    def grid_points(self):
        """Return grid_points.

        See detail at attribute section of this class's docstring.

        """
        return self._grid_points

    @grid_points.setter
    def grid_points(self, grid_points: np.ndarray):
        N = np.prod(self.fft_mesh) * determinant(self.p2s_matrix)
        if grid_points.shape == (N, 3):
            self._grid_points = np.array(grid_points, dtype="double", order="C")
        else:
            raise RuntimeError(
                "Array shape disagreement is found, %s!=%s."
                % ((N, 3), grid_points.shape)
            )

    def run(
        self,
        V_loc_per: np.ndarray,
        V_loc_disps: list[np.ndarray],
        displacements: list[dict],
    ):
        """Calculate dV/du.

        Calculation results are stored in self._dVdu.

        Parameters
        ----------
        V_loc_per : ndarray
            Local potential of perfect supercell
            dtype='complex128'
            shape=(ncdij, nz, ny, nx)
        V_loc_disps : list of ndarrays
            Local potentials of sueprcells with respective displacements
            dtype='complex128'
            shape=(ndisp, ncdij, nz, ny, nx)
        displacements : list of dicts
            Displacements of displaced atoms
            keys of each dict:
                'displacement' : Displacement in Cartesian coordinates
                'number' : Index of displaced atom

        """
        lpi = LocalPotentialInterpolationNUFFT(
            self._fft_mesh,
            self._p2s_matrix,
            self._supercell,
            self._symmetry,
            atom_indices=self._atom_indices,
            nufft=self._nufft,
            finufft_eps=self._finufft_eps,
        )
        self._lattice_points = lpi.lattice_points.copy(order="C")
        self._grid_points = lpi.grid_points.copy(order="C")

        if self._dVdu is None:
            self._allocate_arrays(V_loc_per.shape[0])

        for disp_atom in np.unique([d["number"] for d in displacements]):
            dVs = []
            for i, d in enumerate(displacements):
                if d["number"] == disp_atom:
                    dVs.append(
                        DeltaLocalPotential(V_loc_per, V_loc_disps[i], displacements[i])
                    )
            lpi.delta_Vs = dVs
            for i_atom, _ in enumerate(lpi):  # Run lpi by iterator.next()
                if self._verbose:
                    print(
                        "Computed dV/du by displaced atom %d"
                        % (lpi.atom_indices_returned[i_atom] + 1)
                    )

            indices = []
            for ai in lpi.atom_indices_returned:
                indices.append(np.where(self.atom_indices == ai)[0][0])
            self._dVdu[:, indices, :, :] = lpi.dVdu

    def visualize(self, pcell: PhonopyAtoms, i_atom: int):
        """Visualize dV/du in x, y, z."""
        for i_dir in range(3):
            xyz = "xyz"[i_dir]
            # spin 1 only
            multi = visualize_distribution(
                pcell,
                self._p2s_matrix,
                self._fft_mesh,
                self._grid_points,
                self._dVdu[0, i_atom, i_dir],
                filename="locpot_viz-%03d-%s.dat"
                % (self._atom_indices[i_atom] + 1, xyz),
            )

        if self._verbose:
            msg = (
                "[%d, %d, %d] supercell of primitive cell is "
                "used. dV/du is written in "
                '"locpot_viz-%d-{x,y,z}.dat", in LOCPOT-like '
                "format." % (tuple(multi) + (i_atom + 1,))
            )
            print(
                textwrap.fill(
                    msg, width=70, initial_indent="    ", subsequent_indent="    "
                )
            )

    def _get_dVdu_shape(self):
        num_gp = np.prod(self._fft_mesh) * determinant(self._p2s_matrix)
        return (len(self._atom_indices), 3, num_gp)

    def _allocate_arrays(self, ncdij: int):
        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        shape = (ncdij,) + self._get_dVdu_shape()
        self._dVdu = np.zeros(shape, dtype=dtype, order="C")


def visualize_distribution(
    pcell: PhonopyAtoms,
    p2s_matrix: np.ndarray,
    fft_mesh: Union[int, float, Sequence, np.ndarray],
    grid_points: np.ndarray,
    data: np.ndarray,
    filename: Optional[Union[str, bytes, os.PathLike, io.IOBase]] = None,
) -> np.ndarray:
    """Visualize scalar distribution in space.

    Note
    ----
    multiplicity is the three integers that simply extend primitive cell. This
    extended cell includes the supercell. See more detail in
    _get_multipliticy_for_visualization. This is because the grid is defined
    along primitive cell basis vectors.

    """
    plat = pcell.cell.T
    multiplicity = _get_multipliticy_for_visualization(p2s_matrix, plat)

    M = np.dot(np.linalg.inv(p2s_matrix), np.diag(multiplicity))
    M = np.rint(M).astype(int)
    n = int(np.rint(np.linalg.det(M)))
    d = grid_points.shape[0]

    # Copy grid point coordinates to fill the visualized cell with respect
    # to the basis vectors of local potential cell.
    gp_viz = np.zeros((d * n, 3), dtype="double", order="C")
    snf = SNF3x3(M)
    snf.run()
    P_inv = np.linalg.inv(snf.P)
    diag_A = tuple(np.diagonal(snf.D))

    # Supercell lattice points inside visuzalization cell.
    # Grid points of supercell + the lattice point is concatenated.
    for i, lp in enumerate(np.dot(list(np.ndindex(diag_A)), P_inv.T)):
        gp_viz[(i * d) : ((i + 1) * d), :] = grid_points + lp

    # Grid point coordinates are transformed to those in the visualized cell
    # basis.
    M_inv = np.linalg.inv(M)
    gp_viz = np.dot(gp_viz, M_inv.T)

    # Coordinates are converted to integers.
    multi_to_int = fft_mesh * multiplicity
    gp_viz *= multi_to_int
    gp_viz = np.rint(gp_viz).astype(int) % multi_to_int
    shape = tuple(fft_mesh * multiplicity)[::-1]
    dV_viz = np.zeros(shape, dtype=data.dtype, order="C")
    done = np.zeros(shape, dtype=int, order="C")

    n_ip = len(data)
    for i, index in enumerate(gp_viz):
        dV_viz[tuple(index[::-1])] = data[i % n_ip]
        done[tuple(index[::-1])] = 1

    assert (done == 1).all(), "%d in %d (%d x %d x %d) are done." % (
        done.sum(),
        np.prod(done.shape),
        done.shape[0],
        done.shape[1],
        done.shape[2],
    )

    scell = get_supercell(pcell, np.diag(multiplicity))
    header = "\n".join(get_vasp_structure_lines(scell))
    locpot = get_CHGCAR(dV_viz.real, header)
    with open(filename, "w") as w:
        w.write(locpot)

    return multiplicity


def collect_site_symmetry_operations(
    disp_atom: int, symmetry: Symmetry
) -> tuple[np.ndarray, np.ndarray]:
    """Collect site symmetry operations.

    Collect symmetry operations from a diplaced atom to symmetrically
    equivalent atoms. When the symmetrically equivalent atom is the
    displaced atom, those operations are the site-symmetry
    operations. In the other cases, those operations are bound for
    each symmetrically equivalent atom.

    Returns
    -------
    sitesym_sets : ndarray
        Sets of indices of symmetry operations. Each set contains symmetry
        operations that send the displaced atom to specific one of equivalent
        atoms.
        dtype='int_'
        shape=(equiv_atoms, site_syms)
    equiv_atoms : ndarray
        Indices of symmetrically equivalent atoms
        dtype='int_'
        shape=(equiv_atoms,)

    """
    perms = symmetry.atomic_permutations[:, disp_atom]

    # This gives something like [0, 0, 0, 0, 4, 4, 4, 4, ...].
    map_to_indep_atoms = symmetry.get_map_atoms()
    # This gives something like [0, 1, 2, 3] for disp_atom=0.
    equiv_atoms = np.array(
        [
            i
            for i, x in enumerate(map_to_indep_atoms)
            if map_to_indep_atoms[disp_atom] == x
        ],
        dtype="int_",
    )
    ops_sets = []
    for i_atom in equiv_atoms:
        isyms = np.where(perms == i_atom)[0]
        if len(isyms) > 0:
            ops_sets.append(isyms)

    return np.array(ops_sets, dtype="int_", order="C"), equiv_atoms


def get_displacements_with_rotations(
    rotations: np.ndarray,
    lattice: np.ndarray,
    delta_vals: Union[list[DeltaLocalPotential], list[DeltaDijQij]],
) -> np.ndarray:
    """Rotate displacements by site-symmetry actively.

    Displacements are stored in the following order:

        disps = []
        for calc_disp_dir in disps_of_one_atoms:
            for r in sitesym:
                disps.append(dot(r, calc_disp_dir))

    Returns
    -------
    disps : ndarray
        shape=(calc_disp_dir * sitesyms, 3)

    """
    disps = np.zeros((len(rotations) * len(delta_vals), 3), dtype="double", order="C")
    count = 0
    for delta_V in delta_vals:
        disp = delta_V.displacement["displacement"]
        for r in rotations:
            r_cart = similarity_transformation(lattice, r)
            disps[count] = np.dot(r_cart, disp)
            count += 1
    return disps


def rotate_delta_vals_in_spin_space(
    delta_vals_rotated: list[np.ndarray], r: np.ndarray, lattice: np.ndarray
) -> list[np.ndarray]:
    r"""Take linear combination of delta vals with spin rotation matrix.

    In this method A B A^+ is computed, where A is the 2x2 rotation matrix of
    spin space and B is ``delta_vals_rotated``. What is computed is more
    precisely,

    \sum_{i_s1, i_s2} A_{i_sp, i_s1} B_{i_s1, i_s2} (A^+)_{i_s2, i_s}

    The data order of ``delta_vals_rotated`` is as follows. This method is
    performed on when ncdij=4. This 4 corresponds to 2x2 matrix in spinor
    representation. In VASP covention, the data order is defined as

    cdij=[1, 2, 3, 4] -> [11, 12, 21, 22].

    SpinorRotationMatrices.Delta is a 2x2 numpy array.

    Parameters
    ----------
    delta_vals_rotated : list[np.ndarray]
        Rotated delta local potential or delta-Dijqij.
    lattice : ndarray.
        Basis vectors of supercell in column vectors.
        shape=(3, 3)
    r : ndarray
        Rotation matrix wrt basis vectors (not Cartesian).
        shape=(3, 3).

    """
    srm = SpinorRotationMatrices(r, lattice).run()
    assert srm.Delta.shape == (2, 2)

    delta_vals_spinor_rotated = [np.zeros_like(delta_vals_rotated[0]) for _ in range(4)]
    for i_sp in (0, 1):
        for i_s in (0, 1):
            for i_s1 in (0, 1):
                for i_s2 in (0, 1):
                    d_d = srm.Delta[i_sp, i_s1] * srm.Delta.T.conj()[i_s2, i_s]
                    delta_vals_spinor_rotated[i_sp * 2 + i_s] += (
                        d_d * delta_vals_rotated[i_s1 * 2 + i_s2]
                    )
    return delta_vals_spinor_rotated


def get_grid_points(
    fft_mesh: Union[int, float, Sequence, np.ndarray], p2s_matrix: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    """Return grid points with respect to supercell basis vectors.

    Parameters
    ----------
    fft_mesh: Number of grid points along a, b, c basis vectors of primitve
        cell
    p2s_matrix: An integer matrix that is used to create basis vectors
        of supercell from those of primitive cell

    Returns
    -------
    grid_points : ndarray
        Points to be used to compute <psi|dV_loc|psi'> given as
        coordinates of supercell. Coordinates are given as
        [[x1, y1, z1],
         [x2, y2, z2],
         ...         ]
        dtype='double'
        shape=(prod(fft_mesh)*det(p2s_matrix), 3)
    lattice_points : ndarray
        Lattice points in supercell in primitive cell coordinates
        dtype='int_'
        shape=(det(p2s_matrix), 3)

    """
    # x runs fastest, then y, z. An example is shown belw:
    #
    # In [3]: b, c, a = np.meshgrid(range(3), range(4), range(2))
    # In [4]: a.ravel()
    # Out[4]:
    # array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
    #        0, 1])
    # In [5]: b.ravel()
    # Out[5]:
    # array([0, 0, 1, 1, 2, 2, 0, 0, 1, 1, 2, 2, 0, 0, 1, 1, 2, 2, 0, 0, 1, 1,
    #        2, 2])
    # In [6]: c.ravel()
    # Out[6]:
    # array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3,
    #        3, 3])
    b, c, a = np.meshgrid(range(fft_mesh[1]), range(fft_mesh[2]), range(fft_mesh[0]))
    grid_points = np.multiply(
        np.c_[a.ravel(), b.ravel(), c.ravel()], 1.0 / np.array(fft_mesh)
    )

    # lattice points
    lattice_points, multi = get_lattice_points(p2s_matrix)

    # copy at lattice points
    n = np.prod(fft_mesh)
    gp_super = np.zeros((n * np.prod(multi), 3), dtype="double", order="C")
    for i, lp in enumerate(lattice_points):
        gp_super[(i * n) : ((i + 1) * n), :] = grid_points + lp

    # Transform grid points coordinates from those of primitive cell to those
    # of supercell.
    gp_super[:] = np.dot(gp_super, np.linalg.inv(p2s_matrix).T)

    return gp_super, lattice_points


def _get_multipliticy_for_visualization(
    p2s_matrix: np.ndarray, prim_lattice: np.ndarray
) -> np.ndarray:
    """Find minimum multiplication of primitive cell that can contain supercell.

    Parameters
    ----------
    p2s_matrix : ndarray
        Transformation matrix from primitive cell to supercell.
        shape=(3, 3), dtype='double'
    prim_lattice : ndarray
        Basis vectors of primitive cell given as column vectors.
        shape=(3, 3), dtype='double'

    Returns
    -------
    Multiplicity with respect to primitive cell.

    Note:
        Ls = Lp Ms
        Lv = Ls Mv
        Lv = Lp N
        Ls Mv = Lp N
        Ms Mv = N
        Mv = Ms^-1 N

    """
    plat = prim_lattice  # column vectors
    smat = p2s_matrix
    abc = np.sqrt(np.diagonal(np.dot(plat.T, plat)))
    n = np.ones(3, dtype=int)

    for _ in range(100):
        vmat = np.dot(np.linalg.inv(smat), np.diag(n))
        if (abs(vmat - np.rint(vmat)) < 1e-5).all():
            break
        i_min = np.argmin(abc * n)
        n[i_min] += 1

    return n



"""Implementation of velph-ph_bands-plot."""

import pathlib

import click
import h5py
import matplotlib.pyplot as plt
import numpy as np

from phelel.velph.cli.utils import (
    get_distances_along_BZ_path,
    get_reclat_from_vaspout,
    get_special_points,
)


def plot_ph_bandstructures(
    vaspout_filename: pathlib.Path,
    use_ordinary_frequency: bool = False,
    plot_filename: pathlib.Path = pathlib.Path("ph_bands/ph_bands.pdf"),
):
    """Plot phonon band structure.

    Parameters
    ----------
    vaspout_filename : pathlib.Path
        Filename of vaspout.h5.
    use_ordinary_frequency : bool, optional
        When True, phonon frequency unit becomes ordinary frequency of THz,
        otherwise angular frequency of THz. Default is False.
    plot_filename : pathlib.Path, optional
        File name of band structure plot.

    """
    f = h5py.File(vaspout_filename)
    eigvals = f["results"]["phonons"]["eigenvalues"][:]  # phonon eigenvalues
    if use_ordinary_frequency:
        eigvals /= 2 * np.pi
    omega_max = 1.1 * eigvals.max()

    reclat = get_reclat_from_vaspout(f)
    labels = [
        label.decode("utf-8") for label in f["input"]["qpoints"]["labels_kpoints"][:]
    ]
    nk_per_seg = f["input"]["qpoints"]["number_kpoints"][()]
    kpoint_coords = f["results"]["phonons"]["kpoint_coords"]
    nk_total = len(kpoint_coords)
    k_cart = kpoint_coords @ reclat
    n_segments = nk_total // nk_per_seg
    assert n_segments * nk_per_seg == nk_total
    distances = get_distances_along_BZ_path(nk_total, n_segments, nk_per_seg, k_cart)
    points, labels_at_points = get_special_points(
        labels, distances, n_segments, nk_per_seg, nk_total
    )

    _, ax = plt.subplots()
    ax.plot(distances, eigvals, "ok", markersize=1)
    for x in points[1:-1]:
        ax.vlines(x, 0, omega_max, "k", linewidth=1)
    ax.set_xlim(distances[0], distances[-1])
    ax.set_xticks(points)
    ax.set_xticklabels(labels_at_points)
    ax.set_ylim(0, omega_max)
    if use_ordinary_frequency:
        ax.set_ylabel(r"$\nu$[THz]", fontsize=14)
    else:
        ax.set_ylabel(r"$\omega$[THz]", fontsize=14)

    # Finalize
    plt.rcParams["pdf.fonttype"] = 42
    plt.tight_layout()
    plt.savefig(plot_filename)
    plt.close()

    click.echo(f'Phonon band structure plot was saved in "{plot_filename}".')



"""Utility functions related to data."""

from __future__ import annotations

import numpy as np


def cmplx2real(array: np.ndarray) -> np.ndarray:
    """View numpy array of complex128 by double.

    One complex value is viewed as a pair of real values, i.e.,

        a + 1j * b -> (a, b)

    """
    return array.view("double").reshape(array.shape + (2,))


def real2cmplx(array: np.ndarray) -> np.ndarray:
    """View numpy array of double by complex128.

    One complex value is viewed as a pair of real values, i.e.,

        a + 1j * b <- (a, b)

    """
    dtype = "c%d" % (np.dtype("double").itemsize * 2)
    return array.view(dtype).reshape(array.shape[:-1])



"""Implementation of velph-supercell-generate."""

from __future__ import annotations

import pathlib
import shutil
from typing import Union

import click
import tomli
from phono3py import Phono3py
from phonopy.interface.calculator import write_crystal_structure

import phelel
from phelel import Phelel
from phelel.velph.cli.utils import (
    get_num_digits,
    get_scheduler_dict,
    kspacing_to_mesh,
    write_incar,
    write_kpoints_mesh_mode,
    write_launch_script,
)


def write_supercell_input_files(
    toml_filename: pathlib.Path,
    phelel_yaml_filename: pathlib.Path,
    dir_name: str = "phelel",
) -> None:
    """Generate supercells."""
    if not phelel_yaml_filename.exists():
        click.echo(f'File "{phelel_yaml_filename}" not found.', err=True)
        click.echo('Run "velph phelel init" if necessary.', err=True)
        return None

    phe = phelel.load(phelel_yaml_filename)
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    write_supercells(phe, toml_dict, dir_name=dir_name)
    if phe.phonon_supercell_matrix is not None:
        if "phonon" in toml_dict["vasp"][dir_name]:
            write_phonon_supercells(phe, toml_dict, dir_name=dir_name)
        else:
            print(f'[vasp.{dir_name}.phonon.*] not found in "{toml_filename}"')


def write_supercells(
    phe: Union[Phelel, Phono3py], toml_dict: dict, dir_name: str = "phelel"
):
    """Write VASP input for supercells.

    This is alos used by velph-phono3py-generate.

    """
    kpoints_dict = toml_dict["vasp"][dir_name]["kpoints"]
    if "kspacing" in kpoints_dict:
        symmetry_dataset = kspacing_to_mesh(kpoints_dict, phe.supercell)
        if "symmetry" in toml_dict and "spacegroup_type" in toml_dict["symmetry"]:
            assert (
                symmetry_dataset.international
                == toml_dict["symmetry"]["spacegroup_type"]
            )
    nd = get_num_digits(phe.supercells_with_displacements)

    for i, cell in enumerate(
        [
            phe.supercell,
        ]
        + phe.supercells_with_displacements
    ):
        id_number = f"{i:0{nd}d}"
        disp_dir_name = f"{dir_name}/disp-{id_number}"
        directory = pathlib.Path(disp_dir_name)
        directory.mkdir(parents=True, exist_ok=True)

        _write_vasp_files(
            directory,
            cell,
            toml_dict["vasp"][dir_name]["incar"],
            dir_name,
            kpoints_dict,
        )

        # Scheduler launch script
        if "scheduler" in toml_dict:
            scheduler_dict = get_scheduler_dict(toml_dict, dir_name)
            write_launch_script(scheduler_dict, directory, job_id=id_number)

        click.echo(f'VASP input files were generated in "{disp_dir_name}".')


def write_phonon_supercells(
    phe: Union[Phelel, Phono3py], toml_dict: dict, dir_name: str = "phelel"
):
    """Write VASP input for phonon supercells.

    This is alos used by velph-phono3py-generate.

    """
    kpoints_dict = toml_dict["vasp"][dir_name]["phonon"]["kpoints"]
    nd = get_num_digits(phe.phonon_supercells_with_displacements)

    for i, cell in enumerate(
        [
            phe.phonon_supercell,
        ]
        + phe.phonon_supercells_with_displacements
    ):
        id_number = f"{i:0{nd}d}"
        disp_dir_name = f"{dir_name}/ph-disp-{id_number}"
        directory = pathlib.Path(disp_dir_name)
        directory.mkdir(parents=True, exist_ok=True)

        _write_vasp_files(
            directory,
            cell,
            toml_dict["vasp"][dir_name]["phonon"]["incar"],
            f"{dir_name}.phonon",
            kpoints_dict,
        )

        # Scheduler launch script
        if "scheduler" in toml_dict:
            scheduler_dict = get_scheduler_dict(toml_dict, f"{dir_name}.phonon")
            write_launch_script(scheduler_dict, directory, job_id=id_number)

        click.echo(f'VASP input files were generated in "{disp_dir_name}".')


def _write_vasp_files(directory, cell, toml_incar_dict, dir_name, kpoints_dict):
    # POSCAR
    write_crystal_structure(directory / "POSCAR", cell)

    # INCAR
    write_incar(toml_incar_dict, directory, cell=cell)

    # KPOINTS
    write_kpoints_mesh_mode(
        toml_incar_dict,
        directory,
        f"vasp.{dir_name}.kpoints",
        kpoints_dict,
    )

    # POTCAR
    potcar_path = pathlib.Path("POTCAR")
    if potcar_path.exists():
        shutil.copy2(potcar_path, directory / potcar_path)



"""velph command line tool / velph-ph_bands."""

import pathlib

import click

from phelel.velph.cli import cmd_root
from phelel.velph.cli.ph_bands.generate import (
    write_input_files as write_input_files_ph_bandstructures,
)
from phelel.velph.cli.ph_bands.plot import plot_ph_bandstructures


@cmd_root.group("ph_bands")
@click.help_option("-h", "--help")
def cmd_ph_bands():
    """Choose phonon band structure options."""
    pass


#
# velph ph_bands generate
#
@cmd_ph_bands.command("generate")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str):
    """Generate input files to plot phonon band structure."""
    if not pathlib.Path("POTCAR").exists():
        click.echo('"POTCAR" not found in current directory.')

    write_input_files_ph_bandstructures(pathlib.Path(toml_filename))


#
# velph ph_bands plot
#
@cmd_ph_bands.command("plot")
@click.option(
    "--ordinary-frequency/--angular-frequency",
    "use_ordinary_frequency",
    type=bool,
    default=False,
    help=("Use ordinary frequency instead of angular frequency."),
)
@click.help_option("-h", "--help")
def cmd_plot(use_ordinary_frequency: bool):
    """Plot phonon band structure."""
    vaspout_filename = pathlib.Path("ph_bands/bands/vaspout.h5")
    if not vaspout_filename.exists():
        click.echo(f'"{vaspout_filename}" not found.')
    plot_ph_bandstructures(vaspout_filename, use_ordinary_frequency)



"""velph command line tool / velph-transport."""

from __future__ import annotations

import pathlib
from typing import Optional

import click
import h5py

from phelel.velph.cli.transport import cmd_transport
from phelel.velph.cli.transport.plot.plot_eigenvalues import plot_eigenvalues
from phelel.velph.cli.transport.plot.plot_selfenergy import plot_selfenergy
from phelel.velph.cli.transport.plot.plot_transport import plot_transport


@cmd_transport.group("plot")
@click.help_option("-h", "--help")
def cmd_plot():
    """Choose plot options."""
    pass


@cmd_plot.command("selfenergy")
@click.argument(
    "vaspout_filename",
    nargs=1,
    type=click.Path(),
    default="transport/vaspout.h5",
)
@click.option(
    "--save",
    "save_plot",
    is_flag=bool,
    default=False,
    help=("Save plot to file."),
)
@click.help_option("-h", "--help")
def cmd_plot_selfenergy(vaspout_filename: str, save_plot: bool):
    """Plot self-energy in transports."""
    args = _get_f_h5py_and_plot_filename(
        "selfenergy", vaspout_filename=pathlib.Path(vaspout_filename)
    )
    if args[0] is not None:
        plot_selfenergy(*args, save_plot=save_plot)


@cmd_plot.command("transport")
@click.argument(
    "vaspout_filename",
    nargs=1,
    type=click.Path(),
    default="transport/vaspout.h5",
)
@click.option(
    "--save",
    "save_plot",
    is_flag=bool,
    default=False,
    help=("Save plot to file."),
)
@click.help_option("-h", "--help")
def cmd_plot_transport(vaspout_filename: str, save_plot: bool):
    """Plot transport in transports."""
    args = _get_f_h5py_and_plot_filename(
        "transport", vaspout_filename=pathlib.Path(vaspout_filename)
    )
    if args[0] is not None:
        plot_transport(*args, save_plot=save_plot)


@cmd_plot.command("eigenvalues")
@click.argument(
    "vaspout_filename",
    nargs=1,
    type=click.Path(),
    default="transport/vaspout.h5",
)
@click.option(
    "--cutoff-occupancy",
    nargs=1,
    type=float,
    default=1e-2,
    help=(
        "Cutoff for the occupancy to show eigenvalues in eV. Eigenvalus with "
        "occupances in interval [cutoff_occupancy, 1 - cutoff_occupancy] is "
        "shown. (cutoff_occupancy: float, default=1e-2)"
    ),
)
@click.option(
    "--mu",
    nargs=1,
    type=float,
    default=None,
    help=(
        "Chemical potential in eV unless --tid is specified. "
        "(mu: float, default=None, which means Fermi energy)"
    ),
)
@click.option(
    "--temperature",
    nargs=1,
    type=float,
    default=None,
    help=(
        "Temperature for Fermi-Dirac distribution in K unless --tid is specified. "
        "(temperature: float, default=None, which means 300 K)"
    ),
)
@click.option(
    "--tid",
    nargs=1,
    type=int,
    default=None,
    help=("Index of temperature. (tid: int, default=None)"),
)
@click.help_option("-h", "--help")
def cmd_plot_eigenvalues(
    vaspout_filename: str,
    temperature: float,
    cutoff_occupancy: float,
    mu: Optional[float],
    tid: Optional[int],
):
    """Show eigenvalues in transports."""
    args = _get_f_h5py_and_plot_filename(
        "transport", vaspout_filename=pathlib.Path(vaspout_filename)
    )
    if args[0] is not None:
        plot_eigenvalues(
            args[0],
            tid=tid,
            temperature=temperature,
            cutoff_occupancy=cutoff_occupancy,
            mu=mu,
        )


def _get_f_h5py_and_plot_filename(
    property_name: str,
    vaspout_filename: pathlib.Path = pathlib.Path("transport/vaspout.h5"),
    plot_filename: Optional[pathlib.Path] = None,
) -> tuple[h5py.File, pathlib.Path]:
    if not vaspout_filename.exists():
        click.echo(f'"{vaspout_filename}" (default path) not found.')
        click.echo("Please specify vaspout.h5 file path.")
        return None, None

    if plot_filename is None:
        _plot_filename = vaspout_filename.parent / f"{property_name}.pdf"
    else:
        _plot_filename = plot_filename

    f_h5py = h5py.File(vaspout_filename)

    return f_h5py, _plot_filename
